from __future__ import annotations

import re
from collections.abc import Callable
from contextlib import nullcontext

import pytest

from kasa import Device, DeviceType, KasaException, Module
from tests.conftest import handle_turn_on, turn_on
from tests.device_fixtures import (
    bulb,
    color_bulb,
    non_color_bulb,
    non_variable_temp,
    variable_temp,
)


@bulb
async def test_state_attributes(dev: Device):
    assert "Cloud connection" in dev.state_information
    assert isinstance(dev.state_information["Cloud connection"], bool)


@color_bulb
@turn_on
async def test_hsv(dev: Device, turn_on):
    light = dev.modules.get(Module.Light)
    assert light
    await handle_turn_on(dev, turn_on)
    assert light.has_feature("hsv")

    hue, saturation, brightness = light.hsv
    assert 0 <= hue <= 360
    assert 0 <= saturation <= 100
    assert 0 <= brightness <= 100

    await light.set_hsv(hue=1, saturation=1, value=1)

    await dev.update()
    hue, saturation, brightness = light.hsv
    assert hue == 1
    assert saturation == 1
    assert brightness == 1


@color_bulb
@turn_on
@pytest.mark.parametrize(
    ("hue", "sat", "brightness", "exception_cls", "error"),
    [
        pytest.param(-1, 0, 0, ValueError, "Invalid hue", id="hue out of range"),
        pytest.param(361, 0, 0, ValueError, "Invalid hue", id="hue out of range"),
        pytest.param(
            0.5, 0, 0, TypeError, "Hue must be an integer", id="hue invalid type"
        ),
        pytest.param(
            "foo", 0, 0, TypeError, "Hue must be an integer", id="hue invalid type"
        ),
        pytest.param(
            0, -1, 0, ValueError, "Invalid saturation", id="saturation out of range"
        ),
        pytest.param(
            0, 101, 0, ValueError, "Invalid saturation", id="saturation out of range"
        ),
        pytest.param(
            0,
            0.5,
            0,
            TypeError,
            "Saturation must be an integer",
            id="saturation invalid type",
        ),
        pytest.param(
            0,
            "foo",
            0,
            TypeError,
            "Saturation must be an integer",
            id="saturation invalid type",
        ),
        pytest.param(
            0, 0, -1, ValueError, "Invalid brightness", id="brightness out of range"
        ),
        pytest.param(
            0, 0, 101, ValueError, "Invalid brightness", id="brightness out of range"
        ),
        pytest.param(
            0,
            0,
            0.5,
            TypeError,
            "Brightness must be an integer",
            id="brightness invalid type",
        ),
        pytest.param(
            0,
            0,
            "foo",
            TypeError,
            "Brightness must be an integer",
            id="brightness invalid type",
        ),
    ],
)
async def test_invalid_hsv(
    dev: Device, turn_on, hue, sat, brightness, exception_cls, error
):
    light = dev.modules.get(Module.Light)
    assert light
    await handle_turn_on(dev, turn_on)
    assert light.has_feature("hsv")
    with pytest.raises(exception_cls, match=error):
        await light.set_hsv(hue, sat, brightness)


@color_bulb
@pytest.mark.skip("requires color feature")
async def test_color_state_information(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light
    assert "HSV" in dev.state_information
    assert dev.state_information["HSV"] == light.hsv


@non_color_bulb
async def test_hsv_on_non_color(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light
    assert not light.has_feature("hsv")

    with pytest.raises(KasaException):
        await light.set_hsv(0, 0, 0)
    with pytest.raises(KasaException):
        print(light.hsv)


@variable_temp
@pytest.mark.skip("requires colortemp module")
async def test_variable_temp_state_information(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light
    assert "Color temperature" in dev.state_information
    assert dev.state_information["Color temperature"] == light.color_temp


@variable_temp
@turn_on
async def test_try_set_colortemp(dev: Device, turn_on):
    light = dev.modules.get(Module.Light)
    assert light
    await handle_turn_on(dev, turn_on)
    await light.set_color_temp(2700)
    await dev.update()
    assert light.color_temp == 2700


@variable_temp
async def test_out_of_range_temperature(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light
    with pytest.raises(
        ValueError, match=r"Temperature should be between \d+ and \d+, was 1000"
    ):
        await light.set_color_temp(1000)
    with pytest.raises(
        ValueError, match=r"Temperature should be between \d+ and \d+, was 10000"
    ):
        await light.set_color_temp(10000)


@non_variable_temp
async def test_non_variable_temp(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light
    with pytest.raises(KasaException):
        await light.set_color_temp(2700)

    with pytest.raises(KasaException):
        print(light.color_temp)


@bulb
def test_device_type_bulb(dev: Device):
    assert dev.device_type in {DeviceType.Bulb, DeviceType.LightStrip}


@pytest.mark.parametrize(
    ("attribute", "use_msg", "use_fn"),
    [
        pytest.param(
            "is_color",
            'use has_feature("hsv") instead',
            lambda device, mod: mod.has_feature("hsv"),
            id="is_color",
        ),
        pytest.param(
            "is_dimmable",
            'use has_feature("brightness") instead',
            lambda device, mod: mod.has_feature("brightness"),
            id="is_dimmable",
        ),
        pytest.param(
            "is_variable_color_temp",
            'use has_feature("color_temp") instead',
            lambda device, mod: mod.has_feature("color_temp"),
            id="is_variable_color_temp",
        ),
        pytest.param(
            "has_effects",
            "check `Module.LightEffect in device.modules` instead",
            lambda device, mod: Module.LightEffect in device.modules,
            id="has_effects",
        ),
    ],
)
@bulb
async def test_deprecated_light_is_has_attributes(
    dev: Device, attribute: str, use_msg: str, use_fn: Callable[[Device, Module], bool]
):
    light = dev.modules.get(Module.Light)
    assert light

    msg = f"{attribute} is deprecated, {use_msg}"
    with pytest.deprecated_call(match=(re.escape(msg))):
        result = getattr(light, attribute)

    assert result == use_fn(dev, light)


@bulb
async def test_deprecated_light_valid_temperature_range(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light

    color_temp = light.has_feature("color_temp")
    dep_msg = (
        "valid_temperature_range is deprecated, use "
        'get_feature("color_temp") minimum_value '
        " and maximum_value instead"
    )
    exc_context = pytest.raises(KasaException, match="Color temperature not supported")
    expected_context = nullcontext() if color_temp else exc_context

    with (
        expected_context,
        pytest.deprecated_call(match=(re.escape(dep_msg))),
    ):
        assert light.valid_temperature_range  # type: ignore[attr-defined]



"""Python library supporting TP-Link Smart Home devices.

The communication protocol was reverse engineered by Lubomir Stroetmann and
Tobias Esser in 'Reverse Engineering the TP-Link HS110':
https://www.softscheck.com/en/reverse-engineering-tp-link-hs110/

This library reuses codes and concepts of the TP-Link WiFi SmartPlug Client
at https://github.com/softScheck/tplink-smartplug, developed by Lubomir
Stroetmann which is licensed under the Apache License, Version 2.0.

You may obtain a copy of the license at
http://www.apache.org/licenses/LICENSE-2.0
"""

from __future__ import annotations

import functools
import inspect
import logging
from collections.abc import Callable, Mapping, Sequence
from datetime import datetime, timedelta, tzinfo
from typing import TYPE_CHECKING, Any, cast
from warnings import warn

from ..device import Device, DeviceInfo, WifiNetwork
from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..exceptions import KasaException
from ..feature import Feature
from ..module import Module
from ..modulemapping import ModuleMapping, ModuleName
from ..protocols import BaseProtocol
from .iotmodule import IotModule, merge
from .modules import Emeter

_LOGGER = logging.getLogger(__name__)


def requires_update(f: Callable) -> Any:
    """Indicate that `update` should be called before accessing this method."""  # noqa: D202
    if inspect.iscoroutinefunction(f):

        @functools.wraps(f)
        async def wrapped(*args: Any, **kwargs: Any) -> Any:
            self = args[0]
            if not self._last_update and (
                self._sys_info is None or f.__name__ not in self._sys_info
            ):
                raise KasaException("You need to await update() to access the data")
            return await f(*args, **kwargs)

    else:

        @functools.wraps(f)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            self = args[0]
            if not self._last_update and (
                self._sys_info is None or f.__name__ not in self._sys_info
            ):
                raise KasaException("You need to await update() to access the data")
            return f(*args, **kwargs)

    f.requires_update = True  # type: ignore[attr-defined]
    return wrapped


@functools.lru_cache
def _parse_features(features: str) -> set[str]:
    """Parse features string."""
    return set(features.split(":"))


def _extract_sys_info(info: dict[str, Any]) -> dict[str, Any]:
    """Return the system info structure."""
    sysinfo_default = info.get("system", {}).get("get_sysinfo", {})
    sysinfo_nest = sysinfo_default.get("system", {})

    if len(sysinfo_nest) > len(sysinfo_default) and isinstance(sysinfo_nest, dict):
        return sysinfo_nest
    return sysinfo_default


class IotDevice(Device):
    """Base class for all supported device types.

    You don't usually want to initialize this class manually,
    but either use :class:`Discover` class, or use one of the subclasses:

    * :class:`IotPlug`
    * :class:`IotBulb`
    * :class:`IotStrip`
    * :class:`IotDimmer`
    * :class:`IotLightStrip`

    To initialize, you have to await :func:`update()` at least once.
    This will allow accessing the properties using the exposed properties.

    All changes to the device are done using awaitable methods,
    which will not change the cached values, but you must await update() separately.

    Errors reported by the device are raised as
    :class:`KasaException <kasa.exceptions.KasaException>`,
    and should be handled by the user of the library.

    Examples:
        >>> import asyncio
        >>> dev = IotDevice("127.0.0.1")
        >>> asyncio.run(dev.update())

        All devices provide several informational properties:

        >>> dev.alias
        Bedroom Lamp Plug
        >>> dev.model
        HS110
        >>> dev.rssi
        -71
        >>> dev.mac
        50:C7:BF:00:00:00

        Some information can also be changed programmatically:

        >>> asyncio.run(dev.set_alias("new alias"))
        >>> asyncio.run(dev.set_mac("01:23:45:67:89:ab"))
        >>> asyncio.run(dev.update())
        >>> dev.alias
        new alias
        >>> dev.mac
        01:23:45:67:89:ab

        When initialized using discovery or using a subclass,
        you can check the type of the device:

        >>> dev.is_bulb
        False
        >>> dev.is_strip
        False
        >>> dev.is_plug
        True

        You can also get the hardware and software as a dict,
        or access the full device response:

        >>> dev.hw_info
        {'sw_ver': '1.2.5 Build 171213 Rel.101523',
         'hw_ver': '1.0',
         'mac': '01:23:45:67:89:ab',
         'type': 'IOT.SMARTPLUGSWITCH',
         'hwId': '00000000000000000000000000000000',
         'fwId': '00000000000000000000000000000000',
         'oemId': '00000000000000000000000000000000',
         'dev_name': 'Wi-Fi Smart Plug With Energy Monitoring'}
        >>> dev.sys_info

        All devices can be turned on and off:

        >>> asyncio.run(dev.turn_off())
        >>> asyncio.run(dev.turn_on())
        >>> asyncio.run(dev.update())
        >>> dev.is_on
        True

        Some devices provide energy consumption meter,
        and regular update will already fetch some information:

        >>> dev.has_emeter
        True
        >>> dev.emeter_realtime
        <EmeterStatus power=0.928511 voltage=231.067823 current=0.014937 total=55.139>
        >>> dev.emeter_today
        >>> dev.emeter_this_month

        You can also query the historical data (note that these needs to be awaited),
        keyed with month/day:

        >>> asyncio.run(dev.get_emeter_monthly(year=2016))
        {11: 1.089, 12: 1.582}
        >>> asyncio.run(dev.get_emeter_daily(year=2016, month=11))
        {24: 0.026, 25: 0.109}

    """

    emeter_type = "emeter"

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        """Create a new IotDevice instance."""
        super().__init__(host=host, config=config, protocol=protocol)

        self._sys_info: Any = None  # TODO: this is here to avoid changing tests
        self._supported_modules: dict[str | ModuleName[Module], IotModule] | None = None
        self._legacy_features: set[str] = set()
        self._children: Mapping[str, IotDevice] = {}
        self._modules: dict[str | ModuleName[Module], IotModule] = {}
        self._on_since: datetime | None = None

    @property
    def children(self) -> Sequence[IotDevice]:
        """Return list of children."""
        return list(self._children.values())

    @property
    @requires_update
    def modules(self) -> ModuleMapping[IotModule]:
        """Return the device modules."""
        if TYPE_CHECKING:
            return cast(ModuleMapping[IotModule], self._supported_modules)
        return self._supported_modules

    def add_module(self, name: str | ModuleName[Module], module: IotModule) -> None:
        """Register a module."""
        if name in self._modules:
            _LOGGER.debug("Module %s already registered, ignoring...", name)
            return

        _LOGGER.debug("Adding module %s", module)
        self._modules[name] = module

    def _create_request(
        self,
        target: str,
        cmd: str,
        arg: dict | None = None,
        child_ids: list | None = None,
    ) -> dict:
        if arg is None:
            arg = {}
        request: dict[str, Any] = {target: {cmd: arg}}
        if child_ids is not None:
            request = {"context": {"child_ids": child_ids}, target: {cmd: arg}}

        return request

    def _verify_emeter(self) -> None:
        """Raise an exception if there is no emeter."""
        if not self.has_emeter:
            raise KasaException("Device has no emeter")
        if self.emeter_type not in self._last_update:
            raise KasaException("update() required prior accessing emeter")

    async def _query_helper(
        self,
        target: str,
        cmd: str,
        arg: dict | None = None,
        child_ids: list | None = None,
    ) -> dict:
        """Query device, return results or raise an exception.

        :param target: Target system {system, time, emeter, ..}
        :param cmd: Command to execute
        :param arg: payload dict to be send to the device
        :param child_ids: ids of child devices
        :return: Unwrapped result for the call.
        """
        request = self._create_request(target, cmd, arg, child_ids)

        try:
            response = await self._raw_query(request=request)
        except Exception as ex:
            raise KasaException(f"Communication error on {target}:{cmd}") from ex

        if target not in response:
            raise KasaException(f"No required {target} in response: {response}")

        result = response[target]
        if "err_code" in result and result["err_code"] != 0:
            raise KasaException(f"Error on {target}.{cmd}: {result}")

        if cmd not in result:
            raise KasaException(f"No command in response: {response}")
        result = result[cmd]
        if "err_code" in result and result["err_code"] != 0:
            raise KasaException(f"Error on {target} {cmd}: {result}")

        if "err_code" in result:
            del result["err_code"]

        return result

    @property  # type: ignore
    @requires_update
    def features(self) -> dict[str, Feature]:
        """Return a set of features that the device supports."""
        return self._features

    @property  # type: ignore
    @requires_update
    def has_emeter(self) -> bool:
        """Return True if device has an energy meter."""
        return "ENE" in self._legacy_features

    async def get_sys_info(self) -> dict[str, Any]:
        """Retrieve system information."""
        return await self._query_helper("system", "get_sysinfo")

    async def update(self, update_children: bool = True) -> None:
        """Query the device to update the data.

        Needed for properties that are decorated with `requires_update`.
        """
        req = {}
        req.update(self._create_request("system", "get_sysinfo"))

        # If this is the initial update, check only for the sysinfo
        # This is necessary as some devices crash on unexpected modules
        # See #105, #120, #161
        if not self._last_update:
            _LOGGER.debug("Performing the initial update to obtain sysinfo")
            response = await self.protocol.query(req)
            self._last_update = response
            self._set_sys_info(_extract_sys_info(response))

        if not self._modules:
            await self._initialize_modules()

        await self._modular_update(req)

        self._set_sys_info(_extract_sys_info(self._last_update))
        for module in self._modules.values():
            await module._post_update_hook()

        if not self._features:
            await self._initialize_features()

    async def _initialize_modules(self) -> None:
        """Initialize modules not added in init."""
        if self.has_emeter:
            _LOGGER.debug(
                "The device has emeter, querying its information along sysinfo"
            )
            self.add_module(Module.Energy, Emeter(self, self.emeter_type))

    async def _initialize_features(self) -> None:
        """Initialize common features."""
        self._add_feature(
            Feature(
                self,
                id="state",
                name="State",
                attribute_getter="is_on",
                attribute_setter="set_state",
                type=Feature.Type.Switch,
                category=Feature.Category.Primary,
            )
        )
        self._add_feature(
            Feature(
                device=self,
                id="rssi",
                name="RSSI",
                attribute_getter="rssi",
                icon="mdi:signal",
                unit_getter=lambda: "dBm",
                category=Feature.Category.Debug,
                type=Feature.Type.Sensor,
            )
        )
        # iot strips calculate on_since from the children
        if "on_time" in self._sys_info or self.device_type == Device.Type.Strip:
            self._add_feature(
                Feature(
                    device=self,
                    id="on_since",
                    name="On since",
                    attribute_getter="on_since",
                    icon="mdi:clock",
                    category=Feature.Category.Info,
                    type=Feature.Type.Sensor,
                )
            )

        self._add_feature(
            Feature(
                device=self,
                id="reboot",
                name="Reboot",
                attribute_setter="reboot",
                icon="mdi:restart",
                category=Feature.Category.Debug,
                type=Feature.Type.Action,
            )
        )

        for module in self.modules.values():
            module._initialize_features()
            for module_feat in module._module_features.values():
                self._add_feature(module_feat)

    async def _modular_update(self, req: dict) -> None:
        """Execute an update query."""
        request_list = []
        est_response_size = 1024 if "system" in req else 0
        for module in self._modules.values():
            if not module.is_supported:
                _LOGGER.debug("Module %s not supported, skipping", module)
                continue

            est_response_size += module.estimated_query_response_size
            if est_response_size > self.max_device_response_size:
                request_list.append(req)
                req = {}
                est_response_size = module.estimated_query_response_size

            q = module.query()
            _LOGGER.debug("Adding query for %s: %s", module, q)
            req = merge(req, q)
        request_list.append(req)

        responses = [
            await self.protocol.query(request) for request in request_list if request
        ]

        # Preserve the last update and merge
        # responses on top of it so we remember
        # which modules are not supported, otherwise
        # every other update will query for them
        update: dict = self._last_update.copy() if self._last_update else {}
        for response in responses:
            for k, v in response.items():
                # The same module could have results in different responses
                # i.e. smartlife.iot.common.schedule for Usage and
                # Schedule, so need to call update(**v) here. If a module is
                # not supported the response
                # {'err_code': -1, 'err_msg': 'module not support'}
                # become top level key/values of the response so check for dict
                if isinstance(v, dict):
                    update.setdefault(k, {}).update(**v)
        self._last_update = update

        # IOT modules are added as default but could be unsupported post first update
        if self._supported_modules is None:
            supported = {}
            for module_name, module in self._modules.items():
                if module.is_supported:
                    supported[module_name] = module

            self._supported_modules = supported

    def update_from_discover_info(self, info: dict[str, Any]) -> None:
        """Update state from info from the discover call."""
        self._discovery_info = info
        if "system" in info and (sys_info := info["system"].get("get_sysinfo")):
            self._last_update = info
            self._set_sys_info(sys_info)
        else:
            # This allows setting of some info properties directly
            # from partial discovery info that will then be found
            # by the requires_update decorator
            discovery_model = info["device_model"]
            no_region_model, _, _ = discovery_model.partition("(")
            self._set_sys_info({**info, "model": no_region_model})

    def _set_sys_info(self, sys_info: dict[str, Any]) -> None:
        """Set sys_info."""
        self._sys_info = sys_info
        if features := sys_info.get("feature"):
            self._legacy_features = _parse_features(features)

    @property  # type: ignore
    @requires_update
    def sys_info(self) -> dict[str, Any]:
        """
        Return system information.

        Do not call this function from within the SmartDevice
        class itself as @requires_update will be affected for other properties.
        """
        return self._sys_info  # type: ignore

    @property
    @requires_update
    def model(self) -> str:
        """Returns the device model."""
        if self._last_update:
            return self.device_info.short_name
        return self._sys_info["model"]

    @property  # type: ignore
    def alias(self) -> str | None:
        """Return device name (alias)."""
        sys_info = self._sys_info
        return sys_info.get("alias") if sys_info else None

    async def set_alias(self, alias: str) -> dict:
        """Set the device name (alias)."""
        return await self._query_helper("system", "set_dev_alias", {"alias": alias})

    @property
    @requires_update
    def time(self) -> datetime:
        """Return current time from the device."""
        return self.modules[Module.Time].time

    @property
    @requires_update
    def timezone(self) -> tzinfo:
        """Return the current timezone."""
        return self.modules[Module.Time].timezone

    async def get_time(self) -> datetime:
        """Return current time from the device, if available."""
        msg = "Use `time` property instead, this call will be removed in the future."
        warn(msg, DeprecationWarning, stacklevel=2)
        return self.time

    async def get_timezone(self) -> tzinfo:
        """Return timezone information."""
        msg = (
            "Use `timezone` property instead, this call will be removed in the future."
        )
        warn(msg, DeprecationWarning, stacklevel=2)
        return self.timezone

    @property  # type: ignore
    @requires_update
    def hw_info(self) -> dict:
        """Return hardware information.

        This returns just a selection of sysinfo keys that are related to hardware.
        """
        keys = [
            "sw_ver",
            "hw_ver",
            "mac",
            "mic_mac",
            "type",
            "mic_type",
            "hwId",
            "fwId",
            "oemId",
            "dev_name",
        ]
        sys_info = self._sys_info
        return {key: sys_info[key] for key in keys if key in sys_info}

    @property  # type: ignore
    @requires_update
    def location(self) -> dict:
        """Return geographical location."""
        sys_info = self._sys_info
        loc = {"latitude": None, "longitude": None}

        if "latitude" in sys_info and "longitude" in sys_info:
            loc["latitude"] = sys_info["latitude"]
            loc["longitude"] = sys_info["longitude"]
        elif "latitude_i" in sys_info and "longitude_i" in sys_info:
            loc["latitude"] = sys_info["latitude_i"] / 10000
            loc["longitude"] = sys_info["longitude_i"] / 10000
        else:
            _LOGGER.debug("Unsupported device location.")

        return loc

    @property  # type: ignore
    @requires_update
    def rssi(self) -> int | None:
        """Return WiFi signal strength (rssi)."""
        rssi = self._sys_info.get("rssi")
        return None if rssi is None else int(rssi)

    @property  # type: ignore
    @requires_update
    def mac(self) -> str:
        """Return mac address.

        :return: mac address in hexadecimal with colons, e.g. 01:23:45:67:89:ab
        """
        sys_info = self._sys_info
        mac = sys_info.get("mac", sys_info.get("mic_mac"))
        if not mac:
            raise KasaException(
                "Unknown mac, please submit a bug report with sys_info output."
            )
        mac = mac.replace("-", ":")
        # Format a mac that has no colons (usually from mic_mac field)
        if ":" not in mac:
            mac = ":".join(format(s, "02x") for s in bytes.fromhex(mac))

        return mac

    async def set_mac(self, mac: str) -> dict:
        """Set the mac address.

        :param str mac: mac in hexadecimal with colons, e.g. 01:23:45:67:89:ab
        """
        return await self._query_helper("system", "set_mac_addr", {"mac": mac})

    async def reboot(self, delay: int = 1) -> None:
        """Reboot the device.

        Note that giving a delay of zero causes this to block,
        as the device reboots immediately without responding to the call.
        """
        await self._query_helper("system", "reboot", {"delay": delay})

    async def factory_reset(self) -> None:
        """Reset device back to factory settings.

        Note, this does not downgrade the firmware.
        """
        await self._query_helper("system", "reset")

    async def turn_off(self, **kwargs) -> dict:
        """Turn off the device."""
        raise NotImplementedError("Device subclass needs to implement this.")

    async def turn_on(self, **kwargs) -> dict:
        """Turn device on."""
        raise NotImplementedError("Device subclass needs to implement this.")

    @property  # type: ignore
    @requires_update
    def is_on(self) -> bool:
        """Return True if the device is on."""
        raise NotImplementedError("Device subclass needs to implement this.")

    async def set_state(self, on: bool) -> dict:
        """Set the device state."""
        if on:
            return await self.turn_on()
        else:
            return await self.turn_off()

    @property  # type: ignore
    @requires_update
    def on_since(self) -> datetime | None:
        """Return the time that the device was turned on or None if turned off.

        This returns a cached value if the device reported value difference is under
        five seconds to avoid device-caused jitter.
        """
        if self.is_off or "on_time" not in self._sys_info:
            self._on_since = None
            return None

        on_time = self._sys_info["on_time"]

        on_since = self.time - timedelta(seconds=on_time)
        if not self._on_since or timedelta(
            seconds=0
        ) < on_since - self._on_since > timedelta(seconds=5):
            self._on_since = on_since
        return self._on_since

    @property  # type: ignore
    @requires_update
    def device_id(self) -> str:
        """Return unique ID for the device.

        If not overridden, this is the MAC address of the device.
        Individual sockets on strips will override this.
        """
        return self.mac

    async def wifi_scan(self) -> list[WifiNetwork]:  # noqa: D202
        """Scan for available wifi networks."""

        async def _scan(target: str) -> dict:
            return await self._query_helper(target, "get_scaninfo", {"refresh": 1})

        try:
            info = await _scan("netif")
        except KasaException as ex:
            _LOGGER.debug(
                "Unable to scan using 'netif', retrying with 'softaponboarding': %s", ex
            )
            info = await _scan("smartlife.iot.common.softaponboarding")

        if "ap_list" not in info:
            raise KasaException(f"Invalid response for wifi scan: {info}")

        return [WifiNetwork(**x) for x in info["ap_list"]]

    async def wifi_join(self, ssid: str, password: str, keytype: str = "3") -> dict:  # noqa: D202
        """Join the given wifi network.

        If joining the network fails, the device will return to AP mode after a while.
        """

        async def _join(target: str, payload: dict) -> dict:
            return await self._query_helper(target, "set_stainfo", payload)

        payload = {"ssid": ssid, "password": password, "key_type": int(keytype)}
        try:
            return await _join("netif", payload)
        except KasaException as ex:
            _LOGGER.debug(
                "Unable to join using 'netif', retrying with 'softaponboarding': %s", ex
            )
            return await _join("smartlife.iot.common.softaponboarding", payload)

    @property
    def max_device_response_size(self) -> int:
        """Returns the maximum response size the device can safely construct."""
        return 16 * 1024

    @property
    def internal_state(self) -> Any:
        """Return the internal state of the instance.

        The returned object contains the raw results from the last update call.
        This should only be used for debugging purposes.
        """
        return self._last_update or self._discovery_info

    @staticmethod
    def _get_device_type_from_sys_info(info: dict[str, Any]) -> DeviceType:
        """Find SmartDevice subclass for device described by passed data."""
        if "system" in info.get("system", {}).get("get_sysinfo", {}):
            return DeviceType.Camera

        if "system" not in info or "get_sysinfo" not in info["system"]:
            raise KasaException("No 'system' or 'get_sysinfo' in response")

        sysinfo: dict[str, Any] = _extract_sys_info(info)
        type_: str | None = sysinfo.get("type", sysinfo.get("mic_type"))
        if type_ is None:
            raise KasaException("Unable to find the device type field!")

        if "dev_name" in sysinfo and "Dimmer" in sysinfo["dev_name"]:
            return DeviceType.Dimmer

        if "smartplug" in type_.lower():
            if "children" in sysinfo:
                return DeviceType.Strip
            if (dev_name := sysinfo.get("dev_name")) and "light" in dev_name.lower():
                return DeviceType.WallSwitch
            return DeviceType.Plug

        if "smartbulb" in type_.lower():
            if "length" in sysinfo:  # strips have length
                return DeviceType.LightStrip

            return DeviceType.Bulb

        _LOGGER.warning("Unknown device type %s, falling back to plug", type_)
        return DeviceType.Plug

    @staticmethod
    def _get_device_info(
        info: dict[str, Any], discovery_info: dict[str, Any] | None
    ) -> DeviceInfo:
        """Get model information for a device."""
        sys_info = _extract_sys_info(info)

        # Get model and region info
        region = None
        device_model = sys_info["model"]
        long_name, _, region = device_model.partition("(")
        if region:  # All iot devices have region but just in case
            region = region.replace(")", "")

        # Get other info
        device_family = sys_info.get("type", sys_info.get("mic_type"))
        device_type = IotDevice._get_device_type_from_sys_info(info)
        fw_version_full = sys_info["sw_ver"]
        if " " in fw_version_full:
            firmware_version, firmware_build = fw_version_full.split(" ", maxsplit=1)
        else:
            firmware_version, firmware_build = fw_version_full, None
        auth = bool(discovery_info and ("mgt_encrypt_schm" in discovery_info))

        return DeviceInfo(
            short_name=long_name,
            long_name=long_name,
            brand="kasa",
            device_family=device_family,
            device_type=device_type,
            hardware_version=sys_info["hw_ver"],
            firmware_version=firmware_version,
            firmware_build=firmware_build,
            requires_auth=auth,
            region=region,
        )



"""Provides the current time and timezone information."""

from __future__ import annotations

from datetime import UTC, datetime, tzinfo

from ...exceptions import KasaException
from ...interfaces import Time as TimeInterface
from ..iotmodule import IotModule, merge
from ..iottimezone import get_timezone, get_timezone_index


class Time(IotModule, TimeInterface):
    """Implements the timezone settings."""

    _timezone: tzinfo = UTC

    def query(self) -> dict:
        """Request time and timezone."""
        q = self.query_for_command("get_time")

        merge(q, self.query_for_command("get_timezone"))
        return q

    async def _post_update_hook(self) -> None:
        """Perform actions after a device update."""
        if res := self.data.get("get_timezone"):
            self._timezone = await get_timezone(res.get("index"))

    @property
    def time(self) -> datetime:
        """Return current device time."""
        res = self.data["get_time"]
        time = datetime(
            res["year"],
            res["month"],
            res["mday"],
            res["hour"],
            res["min"],
            res["sec"],
            tzinfo=self.timezone,
        )
        return time

    @property
    def timezone(self) -> tzinfo:
        """Return current timezone."""
        return self._timezone

    async def get_time(self) -> datetime | None:
        """Return current device time."""
        try:
            res = await self.call("get_time")
            return datetime(
                res["year"],
                res["month"],
                res["mday"],
                res["hour"],
                res["min"],
                res["sec"],
                tzinfo=self.timezone,
            )
        except KasaException:
            return None

    async def set_time(self, dt: datetime) -> dict:
        """Set the device time."""
        params = {
            "year": dt.year,
            "month": dt.month,
            "mday": dt.day,
            "hour": dt.hour,
            "min": dt.minute,
            "sec": dt.second,
        }
        if dt.tzinfo:
            index = await get_timezone_index(dt.tzinfo)
            current_index = self.data.get("get_timezone", {}).get("index", -1)
            if current_index != -1 and current_index != index:
                params["index"] = index
                method = "set_timezone"
            else:
                method = "set_time"
        else:
            method = "set_time"
        try:
            return await self.call(method, params)
        except Exception as ex:
            raise KasaException(ex) from ex

    async def get_timezone(self) -> dict:
        """Request timezone information from the device."""
        return await self.call("get_timezone")



from __future__ import annotations

import asyncio
import os
import sys
import warnings
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

# TODO: this and runner fixture could be moved to tests/cli/conftest.py
from asyncclick.testing import CliRunner

from kasa import (
    DeviceConfig,
    SmartProtocol,
)
from kasa.transports.basetransport import BaseTransport

from .device_fixtures import *  # noqa: F403
from .discovery_fixtures import *  # noqa: F403
from .fixtureinfo import fixture_info  # noqa: F401

# Parametrize tests to run with device both on and off
turn_on = pytest.mark.parametrize("turn_on", [True, False])


def load_fixture(foldername, filename):
    """Load a fixture."""
    path = Path(Path(__file__).parent / "fixtures" / foldername / filename)
    with path.open() as fdp:
        return fdp.read()


async def handle_turn_on(dev, turn_on):
    if turn_on:
        await dev.turn_on()
    else:
        await dev.turn_off()


@pytest.fixture
def dummy_protocol():
    """Return a smart protocol instance with a mocking-ready dummy transport."""

    class DummyTransport(BaseTransport):
        @property
        def default_port(self) -> int:
            return -1

        @property
        def credentials_hash(self) -> str:
            return "dummy hash"

        async def send(self, request: str) -> dict:
            return {}

        async def close(self) -> None:
            pass

        async def reset(self) -> None:
            pass

    transport = DummyTransport(config=DeviceConfig(host="127.0.0.123"))
    protocol = SmartProtocol(transport=transport)
    with patch.object(protocol, "BACKOFF_SECONDS_AFTER_TIMEOUT", 0):
        yield protocol


def pytest_configure():
    pytest.fixtures_missing_methods = {}


def pytest_sessionfinish(session, exitstatus):
    if not pytest.fixtures_missing_methods:
        return
    msg = "\n"
    for fixture, methods in sorted(pytest.fixtures_missing_methods.items()):
        method_list = ", ".join(methods)
        msg += f"Fixture {fixture} missing: {method_list}\n"

    warnings.warn(
        UserWarning(msg),
        stacklevel=1,
    )


def pytest_addoption(parser):
    parser.addoption(
        "--ip", action="store", default=None, help="run against device on given ip"
    )
    parser.addoption(
        "--username", action="store", default=None, help="authentication username"
    )
    parser.addoption(
        "--password", action="store", default=None, help="authentication password"
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--ip"):
        print("Testing against fixtures.")
        # pytest_socket doesn't work properly in windows with asyncio
        # fine to disable as other platforms will pickup any issues.
        if sys.platform == "win32":
            for item in items:
                item.add_marker(pytest.mark.enable_socket)
    else:
        print("Running against ip {}".format(config.getoption("--ip")))
        requires_dummy = pytest.mark.skip(
            reason="test requires to be run against dummy data"
        )
        for item in items:
            if "requires_dummy" in item.keywords:
                item.add_marker(requires_dummy)
            else:
                item.add_marker(pytest.mark.enable_socket)


@pytest.fixture(autouse=True, scope="session")
def asyncio_sleep_fixture(request):  # noqa: PT004
    """Patch sleep to prevent tests actually waiting."""
    orig_asyncio_sleep = asyncio.sleep

    async def _asyncio_sleep(*_, **__):
        await orig_asyncio_sleep(0)

    if request.config.getoption("--ip"):
        yield
    else:
        with patch("asyncio.sleep", side_effect=_asyncio_sleep):
            yield


@pytest.fixture(autouse=True, scope="session")
def mock_datagram_endpoint(request):  # noqa: PT004
    """Mock create_datagram_endpoint so it doesn't perform io."""

    async def _create_datagram_endpoint(protocol_factory, *_, **__):
        protocol = protocol_factory()
        transport = MagicMock()
        try:
            return transport, protocol
        finally:
            protocol.connection_made(transport)

    if request.config.getoption("--ip"):
        yield
    else:
        with patch(
            "asyncio.BaseEventLoop.create_datagram_endpoint",
            side_effect=_create_datagram_endpoint,
        ):
            yield


@pytest.fixture
def runner():
    """Runner fixture that unsets the KASA_ environment variables for tests."""
    KASA_VARS = {k: None for k, v in os.environ.items() if k.startswith("KASA_")}
    runner = CliRunner(env=KASA_VARS)

    return runner



"""Typing stub file for ModuleMapping."""

from abc import ABCMeta
from collections.abc import Mapping
from typing import Generic, TypeVar, overload

from .module import Module

__all__ = [
    "ModuleMapping",
    "ModuleName",
]

_ModuleT = TypeVar("_ModuleT", bound=Module, covariant=True)
_ModuleBaseT = TypeVar("_ModuleBaseT", bound=Module, covariant=True)

class ModuleName(Generic[_ModuleT]):
    """Class for typed Module names. At runtime delegated to str."""

    def __init__(self, value: str, /) -> None: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __getitem__(self, index: int) -> str: ...

class ModuleMapping(
    Mapping[ModuleName[_ModuleBaseT] | str, _ModuleBaseT], metaclass=ABCMeta
):
    """Custom dict type to provide better value type hints for Module key types."""

    @overload
    def __getitem__(self, key: ModuleName[_ModuleT], /) -> _ModuleT: ...
    @overload
    def __getitem__(self, key: str, /) -> _ModuleBaseT: ...
    @overload
    def __getitem__(
        self, key: ModuleName[_ModuleT] | str, /
    ) -> _ModuleT | _ModuleBaseT: ...
    @overload  # type: ignore[override]
    def get(self, key: ModuleName[_ModuleT], /) -> _ModuleT | None: ...
    @overload
    def get(self, key: str, /) -> _ModuleBaseT | None: ...
    @overload
    def get(
        self, key: ModuleName[_ModuleT] | str, /
    ) -> _ModuleT | _ModuleBaseT | None: ...

def _test_module_mapping_typing() -> None:
    """Test ModuleMapping overloads work as intended.

    This is tested during the mypy run and needs to be in this file.
    """
    from typing import Any, NewType, assert_type, cast

    from .iot.iotmodule import IotModule
    from .module import Module
    from .smart.smartmodule import SmartModule

    NewCommonModule = NewType("NewCommonModule", Module)
    NewIotModule = NewType("NewIotModule", IotModule)
    NewSmartModule = NewType("NewSmartModule", SmartModule)
    NotModule = NewType("NotModule", list)

    NEW_COMMON_MODULE: ModuleName[NewCommonModule] = ModuleName("NewCommonModule")
    NEW_IOT_MODULE: ModuleName[NewIotModule] = ModuleName("NewIotModule")
    NEW_SMART_MODULE: ModuleName[NewSmartModule] = ModuleName("NewSmartModule")

    # TODO Enable --warn-unused-ignores
    NOT_MODULE: ModuleName[NotModule] = ModuleName("NotModule")  # type: ignore[type-var]  # noqa: F841
    NOT_MODULE_2 = ModuleName[NotModule]("NotModule2")  # type: ignore[type-var]  # noqa: F841

    device_modules: ModuleMapping[Module] = cast(ModuleMapping[Module], {})
    assert_type(device_modules[NEW_COMMON_MODULE], NewCommonModule)
    assert_type(device_modules[NEW_IOT_MODULE], NewIotModule)
    assert_type(device_modules[NEW_SMART_MODULE], NewSmartModule)
    assert_type(device_modules["foobar"], Module)
    assert_type(device_modules[3], Any)  # type: ignore[call-overload]

    assert_type(device_modules.get(NEW_COMMON_MODULE), NewCommonModule | None)
    assert_type(device_modules.get(NEW_IOT_MODULE), NewIotModule | None)
    assert_type(device_modules.get(NEW_SMART_MODULE), NewSmartModule | None)
    assert_type(device_modules.get(NEW_COMMON_MODULE, default=[1, 2]), Any)  # type: ignore[call-overload]

    iot_modules: ModuleMapping[IotModule] = cast(ModuleMapping[IotModule], {})
    smart_modules: ModuleMapping[SmartModule] = cast(ModuleMapping[SmartModule], {})

    assert_type(smart_modules["foobar"], SmartModule)
    assert_type(iot_modules["foobar"], IotModule)

    # Test for covariance
    device_modules_2: ModuleMapping[Module] = iot_modules  # noqa: F841
    device_modules_3: ModuleMapping[Module] = smart_modules  # noqa: F841
    NEW_MODULE: ModuleName[Module] = NEW_SMART_MODULE  # noqa: F841
    NEW_MODULE_2: ModuleName[Module] = NEW_IOT_MODULE  # noqa: F841



"""Light preset module."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING

from mashumaro.config import BaseConfig

from ...exceptions import KasaException
from ...interfaces import LightPreset as LightPresetInterface
from ...interfaces import LightState
from ...json import DataClassJSONMixin
from ...module import Module
from ..iotmodule import IotModule

if TYPE_CHECKING:
    pass

# type ignore can be removed after migration mashumaro:
# error: Signature of "__replace__" incompatible with supertype "LightState"


@dataclass(kw_only=True, repr=False)
class IotLightPreset(DataClassJSONMixin, LightState):  # type: ignore[override]
    """Light configuration preset."""

    class Config(BaseConfig):
        """Config class."""

        omit_none = True

    index: int
    brightness: int

    # These are not available for effect mode presets on light strips
    hue: int | None = None
    saturation: int | None = None
    color_temp: int | None = None

    # Variables for effect mode presets
    custom: int | None = None
    id: str | None = None
    mode: int | None = None


class LightPreset(IotModule, LightPresetInterface):
    """Class for setting light presets."""

    _presets: dict[str, IotLightPreset]
    _preset_list: list[str]

    async def _post_update_hook(self) -> None:
        """Update the internal presets."""
        self._presets = {
            f"Light preset {index + 1}": IotLightPreset.from_dict(vals)
            for index, vals in enumerate(self.data["preferred_state"])
            # Devices may list some light effects along with normal presets but these
            # are handled by the LightEffect module so exclude preferred states with id
            if "id" not in vals
        }
        self._preset_list = [self.PRESET_NOT_SET]
        self._preset_list.extend(self._presets.keys())

    @property
    def preset_list(self) -> list[str]:
        """Return built-in effects list.

        Example:
            ['Off', 'Preset 1', 'Preset 2', ...]
        """
        return self._preset_list

    @property
    def preset_states_list(self) -> Sequence[IotLightPreset]:
        """Return built-in effects list.

        Example:
            ['Off', 'Preset 1', 'Preset 2', ...]
        """
        return list(self._presets.values())

    @property
    def preset(self) -> str:
        """Return current preset name."""
        light = self._device.modules[Module.Light]
        is_color = light.has_feature("hsv")
        is_variable_color_temp = light.has_feature("color_temp")

        brightness = light.brightness
        color_temp = light.color_temp if is_variable_color_temp else None

        h, s = (light.hsv.hue, light.hsv.saturation) if is_color else (None, None)
        for preset_name, preset in self._presets.items():
            if (
                preset.brightness == brightness
                and (preset.color_temp == color_temp or not is_variable_color_temp)
                and (preset.hue == h or not is_color)
                and (preset.saturation == s or not is_color)
            ):
                return preset_name
        return self.PRESET_NOT_SET

    async def set_preset(
        self,
        preset_name: str,
    ) -> dict:
        """Set a light preset for the device."""
        light = self._device.modules[Module.Light]
        if preset_name == self.PRESET_NOT_SET:
            if light.has_feature("hsv"):
                preset = LightState(hue=0, saturation=0, brightness=100)
            else:
                preset = LightState(brightness=100)
        elif (preset := self._presets.get(preset_name)) is None:  # type: ignore[assignment]
            raise ValueError(f"{preset_name} is not a valid preset: {self.preset_list}")

        return await light.set_state(preset)

    @property
    def has_save_preset(self) -> bool:
        """Return True if the device supports updating presets."""
        return True

    async def save_preset(
        self,
        preset_name: str,
        preset_state: LightState,
    ) -> dict:
        """Update the preset with preset_name with the new preset_info."""
        if len(self._presets) == 0:
            raise KasaException("Device does not supported saving presets")
        if preset_name not in self._presets:
            raise ValueError(f"{preset_name} is not a valid preset: {self.preset_list}")

        index = list(self._presets.keys()).index(preset_name)
        state = asdict(preset_state)
        state = {k: v for k, v in state.items() if v is not None}
        state["index"] = index

        return await self.call("set_preferred_state", state)

    def query(self) -> dict:
        """Return the base query."""
        return {}

    @property  # type: ignore
    def _deprecated_presets(self) -> list[IotLightPreset]:
        """Return a list of available bulb setting presets."""
        return [
            IotLightPreset(**vals)
            for vals in self._device.sys_info["preferred_state"]
            if "id" not in vals
        ]

    async def _deprecated_save_preset(self, preset: IotLightPreset) -> dict:
        """Save a setting preset.

        You can either construct a preset object manually, or pass an existing one
        obtained using :func:`presets`.
        """
        if len(self._presets) == 0:
            raise KasaException("Device does not supported saving presets")

        if preset.index >= len(self._presets):
            raise KasaException("Invalid preset index")

        return await self.call("set_preferred_state", preset.to_dict())



import logging
import re

import pytest

from kasa.smart.modules import TemperatureControl
from kasa.smart.modules.temperaturecontrol import ThermostatState

from ...device_fixtures import parametrize, thermostats_smart

temperature = parametrize(
    "has temperature control",
    component_filter="temperature_control",
    protocol_filter={"SMART.CHILD"},
)


@thermostats_smart
@pytest.mark.parametrize(
    ("feature", "type"),
    [
        ("target_temperature", float),
        ("temperature_offset", int),
    ],
)
async def test_temperature_control_features(dev, feature, type):
    """Test that features are registered and work as expected."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]

    prop = getattr(temp_module, feature)
    assert isinstance(prop, type)

    feat = dev.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)

    await feat.set_value(10)
    await dev.update()
    assert feat.value == 10


@thermostats_smart
async def test_set_temperature_turns_heating_on(dev):
    """Test that set_temperature turns heating on."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]

    await temp_module.set_state(False)
    await dev.update()
    assert temp_module.state is False
    assert temp_module.mode is ThermostatState.Off

    await temp_module.set_target_temperature(10)
    await dev.update()
    assert temp_module.state is True
    assert temp_module.mode is ThermostatState.Heating
    assert temp_module.target_temperature == 10


@thermostats_smart
async def test_set_temperature_invalid_values(dev):
    """Test that out-of-bounds temperature values raise errors."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]

    with pytest.raises(
        ValueError, match="Invalid target temperature -1, must be in range"
    ):
        await temp_module.set_target_temperature(-1)

    with pytest.raises(
        ValueError, match="Invalid target temperature 100, must be in range"
    ):
        await temp_module.set_target_temperature(100)


@thermostats_smart
async def test_temperature_offset(dev):
    """Test the temperature offset API."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]
    with pytest.raises(
        ValueError, match=re.escape("Temperature offset must be [-10, 10]")
    ):
        await temp_module.set_temperature_offset(100)

    with pytest.raises(
        ValueError, match=re.escape("Temperature offset must be [-10, 10]")
    ):
        await temp_module.set_temperature_offset(-100)

    await temp_module.set_temperature_offset(5)
    await dev.update()
    assert temp_module.temperature_offset == 5


@thermostats_smart
@pytest.mark.parametrize(
    ("mode", "states", "frost_protection"),
    [
        pytest.param(ThermostatState.Idle, [], False, id="idle has empty"),
        pytest.param(
            ThermostatState.Off,
            ["anything"],
            True,
            id="any state with frost_protection on means off",
        ),
        pytest.param(
            ThermostatState.Heating,
            ["heating"],
            False,
            id="heating is heating",
        ),
        pytest.param(ThermostatState.Unknown, ["invalid"], False, id="unknown state"),
    ],
)
async def test_thermostat_mode(dev, mode, states, frost_protection):
    """Test different thermostat modes."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]

    temp_module.data["frost_protection_on"] = frost_protection
    temp_module.data["trv_states"] = states

    assert temp_module.state is not frost_protection
    assert temp_module.mode is mode


@thermostats_smart
@pytest.mark.parametrize(
    ("mode", "states", "msg"),
    [
        pytest.param(
            ThermostatState.Heating,
            ["heating", "something else"],
            "Got multiple states",
            id="multiple states",
        ),
        pytest.param(
            ThermostatState.Unknown, ["foobar"], "Got unknown state", id="unknown state"
        ),
    ],
)
@pytest.mark.xdist_group(name="caplog")
async def test_thermostat_mode_warnings(dev, mode, states, msg, caplog):
    """Test thermostat modes that should log a warning."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]
    caplog.set_level(logging.WARNING)

    temp_module.data["trv_states"] = states
    assert temp_module.mode is mode
    assert msg in caplog.text


@thermostats_smart
async def test_thermostat_heating_with_low_battery(dev):
    """Test that mode is reported correctly with extra states."""
    temp_module: TemperatureControl = dev.modules["TemperatureControl"]
    temp_module.data["trv_states"] = ["low_battery", "heating"]
    assert temp_module.mode is ThermostatState.Heating



"""Tests for smartcam tamper detection module."""

from __future__ import annotations

from kasa import Device
from kasa.smartcam.smartcammodule import SmartCamModule

from ...device_fixtures import parametrize

tamperdetection = parametrize(
    "has tamper detection",
    component_filter="tamperDetection",
    protocol_filter={"SMARTCAM"},
)


@tamperdetection
async def test_tamperdetection(dev: Device):
    """Test device tamper detection."""
    tamper = dev.modules.get(SmartCamModule.SmartCamTamperDetection)
    assert tamper

    tde_feat = dev.features.get("tamper_detection")
    assert tde_feat

    original_enabled = tamper.enabled

    try:
        await tamper.set_enabled(not original_enabled)
        await dev.update()
        assert tamper.enabled is not original_enabled
        assert tde_feat.value is not original_enabled

        await tamper.set_enabled(original_enabled)
        await dev.update()
        assert tamper.enabled is original_enabled
        assert tde_feat.value is original_enabled

        await tde_feat.set_value(not original_enabled)
        await dev.update()
        assert tamper.enabled is not original_enabled
        assert tde_feat.value is not original_enabled

    finally:
        await tamper.set_enabled(original_enabled)



"""Child lock module."""

from __future__ import annotations

from ...feature import Feature
from ..smartmodule import SmartModule


class ChildLock(SmartModule):
    """Implementation for child lock."""

    REQUIRED_COMPONENT = "button_and_led"
    QUERY_GETTER_NAME = "getChildLockInfo"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                device=self._device,
                id="child_lock",
                name="Child lock",
                container=self,
                attribute_getter="enabled",
                attribute_setter="set_enabled",
                type=Feature.Type.Switch,
                category=Feature.Category.Config,
            )
        )

    @property
    def enabled(self) -> bool:
        """Return True if child lock is enabled."""
        return self.data["child_lock_status"]

    async def set_enabled(self, enabled: bool) -> dict:
        """Set child lock."""
        return await self.call("setChildLockInfo", {"child_lock_status": enabled})



"""Module for cameras."""

from __future__ import annotations

import logging
from datetime import datetime, tzinfo

from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..protocols import BaseProtocol
from .iotdevice import IotDevice

_LOGGER = logging.getLogger(__name__)


class IotCamera(IotDevice):
    """Representation of a TP-Link Camera."""

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        super().__init__(host=host, config=config, protocol=protocol)
        self._device_type = DeviceType.Camera

    @property
    def time(self) -> datetime:
        """Get the camera's time."""
        return datetime.fromtimestamp(self.sys_info["system_time"])

    @property
    def timezone(self) -> tzinfo:
        """Get the camera's timezone."""
        return None  # type: ignore

    @property  # type: ignore
    def is_on(self) -> bool:
        """Return whether device is on."""
        return True



"""Module for dimmers (currently only HS220)."""

from __future__ import annotations

from enum import Enum
from typing import Any

from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..module import Module
from ..protocols import BaseProtocol
from .iotdevice import KasaException, requires_update
from .iotplug import IotPlug
from .modules import AmbientLight, Dimmer, Light, Motion


class ButtonAction(Enum):
    """Button action."""

    NoAction = "none"
    Instant = "instant_on_off"
    Gentle = "gentle_on_off"
    Preset = "customize_preset"


class ActionType(Enum):
    """Button action."""

    DoubleClick = "double_click_action"
    LongPress = "long_press_action"


class FadeType(Enum):
    """Fade on/off setting."""

    FadeOn = "fade_on"
    FadeOff = "fade_off"


class IotDimmer(IotPlug):
    r"""Representation of a TP-Link Smart Dimmer.

    Dimmers work similarly to plugs, but provide also support for
    adjusting the brightness. This class extends :class:`SmartPlug` interface.

    To initialize, you have to await :func:`update()` at least once.
    This will allow accessing the properties using the exposed properties.

    All changes to the device are done using awaitable methods,
    which will not change the cached values,
    but you must await :func:`update()` separately.

    Errors reported by the device are raised as :class:`KasaException`\s,
    and should be handled by the user of the library.

    Examples:
    >>> import asyncio
    >>> dimmer = IotDimmer("192.168.1.105")
    >>> asyncio.run(dimmer.turn_on())
    >>> dimmer.brightness
    25

    >>> asyncio.run(dimmer.set_brightness(50))
    >>> asyncio.run(dimmer.update())
    >>> dimmer.brightness
    50

    Refer to :class:`SmartPlug` for the full API.
    """

    DIMMER_SERVICE = "smartlife.iot.dimmer"

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        super().__init__(host=host, config=config, protocol=protocol)
        self._device_type = DeviceType.Dimmer

    async def _initialize_modules(self) -> None:
        """Initialize modules."""
        await super()._initialize_modules()
        # TODO: need to be verified if it's okay to call these on HS220 w/o these
        # TODO: need to be figured out what's the best approach to detect support
        self.add_module(Module.IotMotion, Motion(self, "smartlife.iot.PIR"))
        self.add_module(Module.IotAmbientLight, AmbientLight(self, "smartlife.iot.LAS"))
        self.add_module(Module.IotDimmer, Dimmer(self, "smartlife.iot.dimmer"))
        self.add_module(Module.Light, Light(self, "light"))

    @property  # type: ignore
    @requires_update
    def _brightness(self) -> int:
        """Return current brightness on dimmers.

        Will return a range between 0 - 100.
        """
        if not self._is_dimmable:
            raise KasaException("Device is not dimmable.")

        sys_info = self.sys_info
        return int(sys_info["brightness"])

    @requires_update
    async def _set_brightness(
        self, brightness: int, *, transition: int | None = None
    ) -> dict:
        """Set the new dimmer brightness level in percentage.

        :param int transition: transition duration in milliseconds.
            Using a transition will cause the dimmer to turn on.
        """
        if not self._is_dimmable:
            raise KasaException("Device is not dimmable.")

        if not isinstance(brightness, int):
            raise ValueError("Brightness must be integer, not of %s.", type(brightness))

        if not 0 <= brightness <= 100:
            raise ValueError(
                f"Invalid brightness value: {brightness} (valid range: 0-100%)"
            )

        # Dimmers do not support a brightness of 0, but bulbs do.
        # Coerce 0 to 1 to maintain the same interface between dimmers and bulbs.
        if brightness == 0:
            brightness = 1

        if transition is not None:
            return await self.set_dimmer_transition(brightness, transition)

        return await self._query_helper(
            self.DIMMER_SERVICE, "set_brightness", {"brightness": brightness}
        )

    async def turn_off(self, *, transition: int | None = None, **kwargs) -> dict:
        """Turn the bulb off.

        :param int transition: transition duration in milliseconds.
        """
        if transition is not None:
            return await self.set_dimmer_transition(brightness=0, transition=transition)

        return await super().turn_off()

    @requires_update
    async def turn_on(self, *, transition: int | None = None, **kwargs) -> dict:
        """Turn the bulb on.

        :param int transition: transition duration in milliseconds.
        """
        if transition is not None:
            return await self.set_dimmer_transition(
                brightness=self._brightness, transition=transition
            )

        return await super().turn_on()

    async def set_dimmer_transition(self, brightness: int, transition: int) -> dict:
        """Turn the bulb on to brightness percentage over transition milliseconds.

        A brightness value of 0 will turn off the dimmer.
        """
        if not isinstance(brightness, int):
            raise TypeError(f"Brightness must be an integer, not {type(brightness)}.")

        if not 0 <= brightness <= 100:
            raise ValueError(
                f"Invalid brightness value: {brightness} (valid range: 0-100%)"
            )

        # If zero set to 1 millisecond
        if transition == 0:
            transition = 1
        if not isinstance(transition, int):
            raise TypeError(f"Transition must be integer, not of {type(transition)}.")
        if transition <= 0:
            raise ValueError(f"Transition value {transition} is not valid.")

        return await self._query_helper(
            self.DIMMER_SERVICE,
            "set_dimmer_transition",
            {"brightness": brightness, "duration": transition},
        )

    @requires_update
    async def get_behaviors(self) -> dict:
        """Return button behavior settings."""
        behaviors = await self._query_helper(
            self.DIMMER_SERVICE, "get_default_behavior", {}
        )
        return behaviors

    @requires_update
    async def set_button_action(
        self, action_type: ActionType, action: ButtonAction, index: int | None = None
    ) -> dict:
        """Set action to perform on button click/hold.

        :param action_type ActionType: whether to control double click or hold action.
        :param action ButtonAction: what should the button do
         (nothing, instant, gentle, change preset)
        :param index int: in case of preset change, the preset to select
        """
        action_type_setter = f"set_{action_type}"

        payload: dict[str, Any] = {"mode": str(action)}
        if index is not None:
            payload["index"] = index

        return await self._query_helper(
            self.DIMMER_SERVICE, action_type_setter, payload
        )

    @requires_update
    async def set_fade_time(self, fade_type: FadeType, time: int) -> dict:
        """Set time for fade in / fade out."""
        fade_type_setter = f"set_{fade_type}_time"
        payload = {"fadeTime": time}

        return await self._query_helper(self.DIMMER_SERVICE, fade_type_setter, payload)

    @property  # type: ignore
    @requires_update
    def _is_dimmable(self) -> bool:
        """Whether the switch supports brightness changes."""
        sys_info = self.sys_info
        return "brightness" in sys_info

    @property
    def _is_variable_color_temp(self) -> bool:
        """Whether the device supports variable color temp."""
        return False

    @property
    def _is_color(self) -> bool:
        """Whether the device supports color."""
        return False



"""Module for base alarm module."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Annotated

from ..module import FeatureAttribute, Module


class Alarm(Module, ABC):
    """Base interface to represent an alarm module."""

    @property
    @abstractmethod
    def alarm_sound(self) -> Annotated[str, FeatureAttribute()]:
        """Return current alarm sound."""

    @abstractmethod
    async def set_alarm_sound(self, sound: str) -> Annotated[dict, FeatureAttribute()]:
        """Set alarm sound.

        See *alarm_sounds* for list of available sounds.
        """

    @property
    @abstractmethod
    def alarm_sounds(self) -> list[str]:
        """Return list of available alarm sounds."""

    @property
    @abstractmethod
    def alarm_volume(self) -> Annotated[int, FeatureAttribute()]:
        """Return alarm volume."""

    @abstractmethod
    async def set_alarm_volume(
        self, volume: int
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set alarm volume."""

    @property
    @abstractmethod
    def alarm_duration(self) -> Annotated[int, FeatureAttribute()]:
        """Return alarm duration."""

    @abstractmethod
    async def set_alarm_duration(
        self, duration: int
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set alarm duration."""

    @property
    @abstractmethod
    def active(self) -> bool:
        """Return true if alarm is active."""

    @abstractmethod
    async def play(
        self,
        *,
        duration: int | None = None,
        volume: int | None = None,
        sound: str | None = None,
    ) -> dict:
        """Play alarm.

        The optional *duration*, *volume*, and *sound* to override the device settings.
        *duration* is in seconds.
        See *alarm_sounds* for the list of sounds available for the device.
        """

    @abstractmethod
    async def stop(self) -> dict:
        """Stop alarm."""



"""Implementation of fan_control module."""

from __future__ import annotations

from typing import Annotated

from ...feature import Feature
from ...interfaces.fan import Fan as FanInterface
from ...module import FeatureAttribute
from ..smartmodule import SmartModule


class Fan(SmartModule, FanInterface):
    """Implementation of fan_control module."""

    REQUIRED_COMPONENT = "fan_control"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="fan_speed_level",
                name="Fan speed level",
                container=self,
                attribute_getter="fan_speed_level",
                attribute_setter="set_fan_speed_level",
                icon="mdi:fan",
                type=Feature.Type.Number,
                range_getter=lambda: (0, 4),
                category=Feature.Category.Primary,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="fan_sleep_mode",
                name="Fan sleep mode",
                container=self,
                attribute_getter="sleep_mode",
                attribute_setter="set_sleep_mode",
                icon="mdi:sleep",
                type=Feature.Type.Switch,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {}

    @property
    def fan_speed_level(self) -> Annotated[int, FeatureAttribute()]:
        """Return fan speed level."""
        return 0 if self.data["device_on"] is False else self.data["fan_speed_level"]

    async def set_fan_speed_level(
        self, level: int
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set fan speed level, 0 for off, 1-4 for on."""
        if level < 0 or level > 4:
            raise ValueError("Invalid level, should be in range 0-4.")
        if level == 0:
            return await self.call("set_device_info", {"device_on": False})
        return await self.call(
            "set_device_info", {"device_on": True, "fan_speed_level": level}
        )

    @property
    def sleep_mode(self) -> Annotated[bool, FeatureAttribute()]:
        """Return sleep mode status."""
        return self.data["fan_sleep_mode_on"]

    async def set_sleep_mode(self, on: bool) -> Annotated[dict, FeatureAttribute()]:
        """Set sleep mode."""
        return await self.call("set_device_info", {"fan_sleep_mode_on": on})

    async def _check_supported(self) -> bool:
        """Is the module available on this device."""
        return "fan_speed_level" in self.data



"""SmartRequest helper classes and functions for new SMART/TAPO devices.

List of known requests with associated parameter classes.

Other requests that are known but not currently implemented
or tested are:

get_child_device_component_list
get_child_device_list
control_child
get_device_running_info - seems to be a subset of get_device_info

get_tss_info
get_raw_dvi
get_homekit_info

fw_download

sync_env
account_sync

device_reset
close_device_ble
heart_beat

"""

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass

_LOGGER = logging.getLogger(__name__)


class SmartRequest:
    """Class to represent a smart protocol request."""

    def __init__(self, method_name: str, params: SmartRequestParams | None = None):
        self.method_name = method_name
        if params:
            self.params = params.to_dict()
        else:
            self.params = None

    def __repr__(self):
        return f"SmartRequest({self.method_name})"

    def to_dict(self):
        """Return the request as a dict suitable for passing to query()."""
        return {self.method_name: self.params}

    @dataclass
    class SmartRequestParams:
        """Base class for Smart request params.

        The to_dict() method of this class omits null values which
        is required by the devices.
        """

        def to_dict(self):
            """Return the params as a dict with values of None ommited."""
            return asdict(
                self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
            )

    @dataclass
    class DeviceOnParams(SmartRequestParams):
        """Get Rules Params."""

        device_on: bool

    @dataclass
    class GetRulesParams(SmartRequestParams):
        """Get Rules Params."""

        start_index: int = 0

    @dataclass
    class GetScheduleRulesParams(SmartRequestParams):
        """Get Rules Params."""

        start_index: int = 0
        schedule_mode: str = ""

    @dataclass
    class GetTriggerLogsParams(SmartRequestParams):
        """Trigger Logs params."""

        page_size: int = 5
        start_id: int = 0

    @dataclass
    class LedStatusParams(SmartRequestParams):
        """LED Status params."""

        led_rule: str | None = None

        @staticmethod
        def from_bool(state: bool):
            """Set the led_rule from the state."""
            rule = "always" if state else "never"
            return SmartRequest.LedStatusParams(led_rule=rule)

    @dataclass
    class LightInfoParams(SmartRequestParams):
        """LightInfo params."""

        brightness: int | None = None
        color_temp: int | None = None
        hue: int | None = None
        saturation: int | None = None

    @dataclass
    class DynamicLightEffectParams(SmartRequestParams):
        """LightInfo params."""

        enable: bool
        id: str | None = None

    @dataclass
    class GetCleanAttrParams(SmartRequestParams):
        """CleanAttr params.

        Decides which cleaning settings are requested
        """

        #: type can be global or pose
        type: str = "global"

    @staticmethod
    def get_raw_request(
        method: str, params: SmartRequestParams | None = None
    ) -> SmartRequest:
        """Send a raw request to the device."""
        return SmartRequest(method, params)

    @staticmethod
    def component_nego() -> SmartRequest:
        """Get quick setup component info."""
        return SmartRequest("component_nego")

    @staticmethod
    def get_device_info() -> SmartRequest:
        """Get device info."""
        return SmartRequest("get_device_info")

    @staticmethod
    def get_device_usage() -> SmartRequest:
        """Get device usage."""
        return SmartRequest("get_device_usage")

    @staticmethod
    def device_info_list(ver_code) -> list[SmartRequest]:
        """Get device info list."""
        if ver_code == 1:
            return [SmartRequest.get_device_info()]
        return [
            SmartRequest.get_device_info(),
            SmartRequest.get_device_usage(),
            SmartRequest.get_auto_update_info(),
        ]

    @staticmethod
    def get_auto_update_info() -> SmartRequest:
        """Get auto update info."""
        return SmartRequest("get_auto_update_info")

    @staticmethod
    def firmware_info_list() -> list[SmartRequest]:
        """Get info list."""
        return [
            SmartRequest.get_raw_request("get_fw_download_state"),
            SmartRequest.get_raw_request("get_latest_fw"),
        ]

    @staticmethod
    def qs_component_nego() -> SmartRequest:
        """Get quick setup component info."""
        return SmartRequest("qs_component_nego")

    @staticmethod
    def get_device_time() -> SmartRequest:
        """Get device time."""
        return SmartRequest("get_device_time")

    @staticmethod
    def get_child_device_list() -> SmartRequest:
        """Get child device list."""
        return SmartRequest("get_child_device_list")

    @staticmethod
    def get_child_device_component_list() -> SmartRequest:
        """Get child device component list."""
        return SmartRequest("get_child_device_component_list")

    @staticmethod
    def get_wireless_scan_info(
        params: GetRulesParams | None = None,
    ) -> SmartRequest:
        """Get wireless scan info."""
        return SmartRequest(
            "get_wireless_scan_info", params or SmartRequest.GetRulesParams()
        )

    @staticmethod
    def get_schedule_rules(params: GetRulesParams | None = None) -> SmartRequest:
        """Get schedule rules."""
        return SmartRequest(
            "get_schedule_rules", params or SmartRequest.GetScheduleRulesParams()
        )

    @staticmethod
    def get_next_event(params: GetRulesParams | None = None) -> SmartRequest:
        """Get next scheduled event."""
        return SmartRequest("get_next_event", params or SmartRequest.GetRulesParams())

    @staticmethod
    def schedule_info_list() -> list[SmartRequest]:
        """Get schedule info list."""
        return [
            SmartRequest.get_schedule_rules(),
            SmartRequest.get_next_event(),
        ]

    @staticmethod
    def get_countdown_rules(params: GetRulesParams | None = None) -> SmartRequest:
        """Get countdown rules."""
        return SmartRequest(
            "get_countdown_rules", params or SmartRequest.GetRulesParams()
        )

    @staticmethod
    def get_antitheft_rules(params: GetRulesParams | None = None) -> SmartRequest:
        """Get antitheft rules."""
        return SmartRequest(
            "get_antitheft_rules", params or SmartRequest.GetRulesParams()
        )

    @staticmethod
    def get_led_info(params: LedStatusParams | None = None) -> SmartRequest:
        """Get led info."""
        return SmartRequest("get_led_info", params or SmartRequest.LedStatusParams())

    @staticmethod
    def get_auto_off_config(params: GetRulesParams | None = None) -> SmartRequest:
        """Get auto off config."""
        return SmartRequest(
            "get_auto_off_config", params or SmartRequest.GetRulesParams()
        )

    @staticmethod
    def get_delay_action_info() -> SmartRequest:
        """Get delay action info."""
        return SmartRequest("get_delay_action_info")

    @staticmethod
    def auto_off_list() -> list[SmartRequest]:
        """Get energy usage."""
        return [
            SmartRequest.get_auto_off_config(),
            SmartRequest.get_delay_action_info(),  # May not live here
        ]

    @staticmethod
    def get_energy_usage() -> SmartRequest:
        """Get energy usage."""
        return SmartRequest("get_energy_usage")

    @staticmethod
    def energy_monitoring_list() -> list[SmartRequest]:
        """Get energy usage."""
        return [
            SmartRequest("get_energy_usage"),
            SmartRequest("get_emeter_data"),
            SmartRequest("get_emeter_vgain_igain"),
            SmartRequest.get_raw_request("get_electricity_price_config"),
        ]

    @staticmethod
    def get_current_power() -> SmartRequest:
        """Get current power."""
        return SmartRequest("get_current_power")

    @staticmethod
    def power_protection_list() -> list[SmartRequest]:
        """Get power protection info list."""
        return [
            SmartRequest.get_current_power(),
            SmartRequest.get_raw_request("get_max_power"),
            SmartRequest.get_raw_request("get_protection_power"),
        ]

    @staticmethod
    def get_preset_rules(params: GetRulesParams | None = None) -> SmartRequest:
        """Get preset rules."""
        return SmartRequest("get_preset_rules", params or SmartRequest.GetRulesParams())

    @staticmethod
    def get_on_off_gradually_info(
        params: SmartRequestParams | None = None,
    ) -> SmartRequest:
        """Get preset rules."""
        return SmartRequest(
            "get_on_off_gradually_info", params or SmartRequest.SmartRequestParams()
        )

    @staticmethod
    def get_auto_light_info() -> SmartRequest:
        """Get auto light info."""
        return SmartRequest("get_auto_light_info")

    @staticmethod
    def get_dynamic_light_effect_rules(
        params: GetRulesParams | None = None,
    ) -> SmartRequest:
        """Get dynamic light effect rules."""
        return SmartRequest(
            "get_dynamic_light_effect_rules", params or SmartRequest.GetRulesParams()
        )

    @staticmethod
    def set_device_on(params: DeviceOnParams) -> SmartRequest:
        """Set device on state."""
        return SmartRequest("set_device_info", params)

    @staticmethod
    def set_light_info(params: LightInfoParams) -> SmartRequest:
        """Set color temperature."""
        return SmartRequest("set_device_info", params)

    @staticmethod
    def set_dynamic_light_effect_rule_enable(
        params: DynamicLightEffectParams,
    ) -> SmartRequest:
        """Enable dynamic light effect rule."""
        return SmartRequest("set_dynamic_light_effect_rule_enable", params)

    @staticmethod
    def get_component_info_requests(component_nego_response) -> list[SmartRequest]:
        """Get a list of requests based on the component info response."""
        request_list: list[SmartRequest] = []
        for component in component_nego_response["component_list"]:
            if (
                requests := get_component_requests(
                    component["id"], int(component["ver_code"])
                )
            ) is not None:
                request_list.extend(requests)
        return request_list

    @staticmethod
    def _create_request_dict(
        smart_request: SmartRequest | list[SmartRequest],
    ) -> dict:
        """Create request dict to be passed to SmartProtocol.query()."""
        if isinstance(smart_request, list):
            request = {}
            for sr in smart_request:
                request[sr.method_name] = sr.params
        else:
            request = smart_request.to_dict()
        return request


def get_component_requests(component_id, ver_code):
    """Get the requests supported by the component and version."""
    if (cr := COMPONENT_REQUESTS.get(component_id)) is None:
        return None
    if callable(cr):
        return SmartRequest._create_request_dict(cr(ver_code))
    return SmartRequest._create_request_dict(cr)


COMPONENT_REQUESTS = {
    "device": SmartRequest.device_info_list,
    "firmware": SmartRequest.firmware_info_list(),
    "quick_setup": [SmartRequest.qs_component_nego()],
    "inherit": [SmartRequest.get_raw_request("get_inherit_info")],
    "time": [SmartRequest.get_device_time()],
    "wireless": [SmartRequest.get_wireless_scan_info()],
    "schedule": SmartRequest.schedule_info_list(),
    "countdown": [SmartRequest.get_countdown_rules()],
    "antitheft": [SmartRequest.get_antitheft_rules()],
    "account": [],
    "synchronize": [],  # sync_env
    "sunrise_sunset": [],  # for schedules
    "led": [SmartRequest.get_led_info()],
    "cloud_connect": [SmartRequest.get_raw_request("get_connect_cloud_state")],
    "iot_cloud": [],
    "device_local_time": [],
    "default_states": [],  # in device_info
    "auto_off": [SmartRequest.get_auto_off_config()],
    "localSmart": [],
    "energy_monitoring": SmartRequest.energy_monitoring_list(),
    "power_protection": SmartRequest.power_protection_list(),
    "current_protection": [],  # overcurrent in device_info
    "matter": [SmartRequest.get_raw_request("get_matter_setup_info")],
    "preset": [SmartRequest.get_preset_rules()],
    "brightness": [],  # in device_info
    "color": [],  # in device_info
    "color_temperature": [],  # in device_info
    "auto_light": [SmartRequest.get_auto_light_info()],
    "light_effect": [SmartRequest.get_dynamic_light_effect_rules()],
    "bulb_quick_control": [],
    "on_off_gradually": [SmartRequest.get_on_off_gradually_info()],
    "light_strip": [],
    "light_strip_lighting_effect": [
        SmartRequest.get_raw_request("get_lighting_effect")
    ],
    "music_rhythm": [],  # music_rhythm_enable in device_info
    "segment": [SmartRequest.get_raw_request("get_device_segment")],
    "segment_effect": [SmartRequest.get_raw_request("get_segment_effect_rule")],
    "device_load": [SmartRequest.get_raw_request("get_device_load_info")],
    "child_quick_setup": [
        SmartRequest.get_raw_request("get_support_child_device_category")
    ],
    "alarm": [
        SmartRequest.get_raw_request("get_support_alarm_type_list"),
        SmartRequest.get_raw_request("get_alarm_configure"),
    ],
    "alarm_logs": [SmartRequest.get_raw_request("get_alarm_triggers")],
    "trigger_log": [
        SmartRequest.get_raw_request(
            "get_trigger_logs", SmartRequest.GetTriggerLogsParams()
        )
    ],
    "temp_humidity_record": [SmartRequest.get_raw_request("get_temp_humidity_records")],
    "double_click": [SmartRequest.get_raw_request("get_double_click_info")],
    "child_device": [
        SmartRequest.get_raw_request("get_child_device_list"),
        SmartRequest.get_raw_request("get_child_device_component_list"),
    ],
    "control_child": [],
    "homekit": [SmartRequest.get_raw_request("get_homekit_info")],
    "dimmer_calibration": [],
    "fan_control": [],
    "overheat_protection": [],
    # Vacuum components
    "clean": [
        SmartRequest.get_raw_request("getCarpetClean"),
        SmartRequest.get_raw_request("getCleanRecords"),
        SmartRequest.get_raw_request("getVacStatus"),
        SmartRequest.get_raw_request("getAreaUnit"),
        SmartRequest.get_raw_request("getCleanInfo"),
        SmartRequest.get_raw_request("getCleanStatus"),
        SmartRequest("getCleanAttr", SmartRequest.GetCleanAttrParams()),
    ],
    "battery": [SmartRequest.get_raw_request("getBatteryInfo")],
    "consumables": [SmartRequest.get_raw_request("getConsumablesInfo")],
    "direction_control": [],
    "button_and_led": [SmartRequest.get_raw_request("getChildLockInfo")],
    "speaker": [
        SmartRequest.get_raw_request("getSupportVoiceLanguage"),
        SmartRequest.get_raw_request("getCurrentVoiceLanguage"),
        SmartRequest.get_raw_request("getVolume"),
    ],
    "map": [
        SmartRequest.get_raw_request("getMapInfo"),
        SmartRequest.get_raw_request("getMapData"),
    ],
    "auto_change_map": [SmartRequest.get_raw_request("getAutoChangeMap")],
    "dust_bucket": [
        SmartRequest.get_raw_request("getAutoDustCollection"),
        SmartRequest.get_raw_request("getDustCollectionInfo"),
    ],
    "mop": [SmartRequest.get_raw_request("getMopState")],
    "do_not_disturb": [SmartRequest.get_raw_request("getDoNotDisturb")],
    "charge_pose_clean": [],
    "continue_breakpoint_sweep": [],
    "goto_point": [],
}



import datetime
from unittest.mock import Mock

from kasa.iot.modules import Usage


def test_usage_convert_stat_data():
    usage = Usage(None, module="usage")

    test_data = []
    assert usage._convert_stat_data(test_data, "day") == {}

    test_data = [
        {"year": 2016, "month": 5, "day": 2, "time": 20},
        {"year": 2016, "month": 5, "day": 4, "time": 30},
    ]
    d = usage._convert_stat_data(test_data, "day")
    assert len(d) == len(test_data)
    assert isinstance(d, dict)
    k, v = d.popitem()
    assert isinstance(k, int)
    assert isinstance(v, int)
    assert k == 4
    assert v == 30


def test_usage_today():
    """Test fetching the usage for today.

    This test uses inline data since the fixtures
    will not have data for the current day.
    """
    emeter_data = {
        "get_daystat": {
            "day_list": [],
            "err_code": 0,
        }
    }

    class MockUsage(Usage):
        @property
        def data(self):
            return emeter_data

    usage = MockUsage(Mock(), "usage")
    assert usage.usage_today is None
    now = datetime.datetime.now()
    emeter_data["get_daystat"]["day_list"].extend(
        [
            {"day": now.day - 1, "time": 200, "month": now.month - 1, "year": now.year},
            {"day": now.day, "time": 500, "month": now.month, "year": now.year},
            {"day": now.day + 1, "time": 100, "month": now.month + 1, "year": now.year},
        ]
    )
    assert usage.usage_today == 500


def test_usage_this_month():
    """Test fetching the usage for this month.

    This test uses inline data since the fixtures
    will not have data for the current month.
    """
    emeter_data = {
        "get_monthstat": {
            "month_list": [],
            "err_code": 0,
        }
    }

    class MockUsage(Usage):
        @property
        def data(self):
            return emeter_data

    usage = MockUsage(Mock(), "usage")
    assert usage.usage_this_month is None
    now = datetime.datetime.now()
    emeter_data["get_monthstat"]["month_list"].extend(
        [
            {"time": 200, "month": now.month - 1, "year": now.year},
            {"time": 500, "month": now.month, "year": now.year},
            {"time": 100, "month": now.month + 1, "year": now.year},
        ]
    )
    assert usage.usage_this_month == 500



"""Module for light effects."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..effects import EFFECT_MAPPING, EFFECT_NAMES, SmartLightEffect
from ..smartmodule import Module, SmartModule, allow_update_after

if TYPE_CHECKING:
    from ..smartdevice import SmartDevice


class LightStripEffect(SmartModule, SmartLightEffect):
    """Implementation of dynamic light effects."""

    REQUIRED_COMPONENT = "light_strip_lighting_effect"

    def __init__(self, device: SmartDevice, module: str) -> None:
        super().__init__(device, module)
        effect_list = [self.LIGHT_EFFECTS_OFF]
        effect_list.extend(EFFECT_NAMES)
        self._effect_list = effect_list
        self._effect_mapping = EFFECT_MAPPING

    @property
    def name(self) -> str:
        """Name of the module.

        By default smart modules are keyed in the module mapping by class name.
        The name is overriden here as this module implements the same common interface
        as the bulb light_effect and the assumption is a device only supports one
        or the other.

        """
        return "LightEffect"

    @property
    def effect(self) -> str:
        """Return effect name."""
        eff = self.data["lighting_effect"]
        name = eff["name"]
        # When devices are unpaired effect name is softAP which is not in our list
        if eff["enable"] and name in self._effect_list:
            return name
        if eff["enable"] and eff["custom"]:
            return name or self.LIGHT_EFFECTS_UNNAMED_CUSTOM
        return self.LIGHT_EFFECTS_OFF

    @property
    def is_active(self) -> bool:
        """Return if effect is active."""
        eff = self.data["lighting_effect"]
        # softAP has enable=1, but brightness 0 which fails on tests
        return bool(eff["enable"]) and eff["name"] in self._effect_list

    @property
    def brightness(self) -> int:
        """Return effect brightness."""
        eff = self.data["lighting_effect"]
        return eff["brightness"]

    async def set_brightness(
        self, brightness: int, *, transition: int | None = None
    ) -> dict:
        """Set effect brightness."""
        if brightness <= 0:
            return await self.set_effect(self.LIGHT_EFFECTS_OFF)

        # Need to pass bAdjusted to keep the existing effect running
        eff = {"brightness": brightness, "bAdjusted": True}
        return await self.set_custom_effect(eff)

    @property
    def effect_list(self) -> list[str]:
        """Return built-in effects list.

        Example:
            ['Aurora', 'Bubbling Cauldron', ...]
        """
        return self._effect_list

    @allow_update_after
    async def set_effect(
        self,
        effect: str,
        *,
        brightness: int | None = None,
        transition: int | None = None,
    ) -> dict:
        """Set an effect on the device.

        If brightness or transition is defined,
        its value will be used instead of the effect-specific default.

        See :meth:`effect_list` for available effects,
        or use :meth:`set_custom_effect` for custom effects.

        :param str effect: The effect to set
        :param int brightness: The wanted brightness
        :param int transition: The wanted transition time
        """
        brightness_module = self._device.modules[Module.Brightness]
        if effect == self.LIGHT_EFFECTS_OFF:
            if self.effect in self._effect_mapping:
                # TODO: We could query get_lighting_effect here to
                # get the custom effect although not sure how to find
                # custom effects
                effect_dict = self._effect_mapping[self.effect]
            else:
                effect_dict = self._effect_mapping["Aurora"]
            effect_dict = {**effect_dict}
            effect_dict["enable"] = 0
            return await self.set_custom_effect(effect_dict)

        if effect not in self._effect_mapping:
            raise ValueError(f"The effect {effect} is not a built in effect.")
        else:
            effect_dict = self._effect_mapping[effect]
            effect_dict = {**effect_dict}

        # Use explicitly given brightness
        if brightness is not None:
            effect_dict["brightness"] = brightness
        # Fall back to brightness reported by the brightness module
        elif brightness_module.brightness:
            effect_dict["brightness"] = brightness_module.brightness

        if transition is not None:
            effect_dict["transition"] = transition

        return await self.set_custom_effect(effect_dict)

    @allow_update_after
    async def set_custom_effect(
        self,
        effect_dict: dict,
    ) -> dict:
        """Set a custom effect on the device.

        :param str effect_dict: The custom effect dict to set
        """
        return await self.call(
            "set_lighting_effect",
            effect_dict,
        )

    @property
    def has_custom_effects(self) -> bool:
        """Return True if the device supports setting custom effects."""
        return True

    def query(self) -> dict:
        """Return the base query."""
        return {}

    @property  # type: ignore
    def _deprecated_effect(self) -> dict:
        """Return effect state.

        Example:
            {'brightness': 50,
             'custom': 0,
             'enable': 0,
             'id': '',
             'name': ''}
        """
        # LightEffectModule returns the current effect name
        # so return the dict here for backwards compatibility
        return self.data["lighting_effect"]

    @property  # type: ignore
    def _deprecated_effect_list(self) -> list[str] | None:
        """Return built-in effects list.

        Example:
            ['Aurora', 'Bubbling Cauldron', ...]
        """
        # LightEffectModule returns effect names along with a LIGHT_EFFECTS_OFF value
        # so return the original effect names here for backwards compatibility
        return EFFECT_NAMES



"""Implementation for child device setup.

This module allows pairing and disconnecting child devices.
"""

from __future__ import annotations

import asyncio
import logging

from ...feature import Feature
from ...interfaces.childsetup import ChildSetup as ChildSetupInterface
from ..smartmodule import SmartModule

_LOGGER = logging.getLogger(__name__)


class ChildSetup(SmartModule, ChildSetupInterface):
    """Implementation for child device setup."""

    REQUIRED_COMPONENT = "child_quick_setup"
    QUERY_GETTER_NAME = "get_support_child_device_category"
    _categories: list[str] = []

    # Supported child device categories will hardly ever change
    MINIMUM_UPDATE_INTERVAL_SECS = 60 * 60 * 24

    def _initialize_features(self) -> None:
        """Initialize features."""
        self._add_feature(
            Feature(
                self._device,
                id="pair",
                name="Pair",
                container=self,
                attribute_setter="pair",
                category=Feature.Category.Config,
                type=Feature.Type.Action,
            )
        )

    async def _post_update_hook(self) -> None:
        self._categories = [
            cat["category"] for cat in self.data["device_category_list"]
        ]

    @property
    def supported_categories(self) -> list[str]:
        """Supported child device categories."""
        return self._categories

    async def pair(self, *, timeout: int = 10) -> list[dict]:
        """Scan for new devices and pair them."""
        await self.call("begin_scanning_child_device")

        _LOGGER.info("Waiting %s seconds for discovering new devices", timeout)
        await asyncio.sleep(timeout)
        detected = await self._get_detected_devices()

        if not detected["child_device_list"]:
            _LOGGER.warning(
                "No devices found, make sure to activate pairing "
                "mode on the devices to be added."
            )
            return []

        _LOGGER.info(
            "Discovery done, found %s devices: %s",
            len(detected["child_device_list"]),
            detected,
        )

        return await self._add_devices(detected)

    async def unpair(self, device_id: str) -> dict:
        """Remove device from the hub."""
        _LOGGER.info("Going to unpair %s from %s", device_id, self)

        payload = {"child_device_list": [{"device_id": device_id}]}
        res = await self.call("remove_child_device_list", payload)
        await self._device.update()
        return res

    async def _add_devices(self, devices: dict) -> list[dict]:
        """Add devices based on get_detected_device response.

        Pass the output from :ref:_get_detected_devices: as a parameter.
        """
        await self.call("add_child_device_list", devices)

        await self._device.update()

        successes = []
        for detected in devices["child_device_list"]:
            device_id = detected["device_id"]

            result = "not added"
            if device_id in self._device._children:
                result = "added"
                successes.append(detected)

            msg = f"{detected['device_model']} - {device_id} - {result}"
            _LOGGER.info("Added child to %s: %s", self._device.host, msg)

        return successes

    async def _get_detected_devices(self) -> dict:
        """Return list of devices detected during scanning."""
        param = {"scan_list": self.data["device_category_list"]}
        res = await self.call("get_scan_child_device_list", param)
        _LOGGER.debug("Scan status: %s", res)
        return res["get_scan_child_device_list"]



"""Module for io device timezone lookups."""

from __future__ import annotations

import logging
from datetime import datetime, timedelta, tzinfo
from typing import cast
from zoneinfo import ZoneInfo

from ..cachedzoneinfo import CachedZoneInfo

_LOGGER = logging.getLogger(__name__)


async def get_timezone(index: int) -> tzinfo:
    """Get the timezone from the index."""
    if index > 109:
        _LOGGER.error(
            "Unexpected index %s not configured as a timezone, defaulting to UTC", index
        )
        return await CachedZoneInfo.get_cached_zone_info("Etc/UTC")

    name = TIMEZONE_INDEX[index]
    return await CachedZoneInfo.get_cached_zone_info(name)


async def get_timezone_index(tzone: tzinfo) -> int:
    """Return the iot firmware index for a valid IANA timezone key."""
    if isinstance(tzone, ZoneInfo):
        name = tzone.key
        rev = {val: key for key, val in TIMEZONE_INDEX.items()}
        if name in rev:
            return rev[name]

    for i in range(110):
        if _is_same_timezone(tzone, await get_timezone(i)):
            return i
    raise ValueError("Device does not support timezone %s", name)


async def get_matching_timezones(tzone: tzinfo) -> list[str]:
    """Return the iot firmware index for a valid IANA timezone key."""
    matches = []
    if isinstance(tzone, ZoneInfo):
        name = tzone.key
        vals = {val for val in TIMEZONE_INDEX.values()}
        if name in vals:
            matches.append(name)

    for i in range(110):
        fw_tz = await get_timezone(i)
        if _is_same_timezone(tzone, fw_tz):
            match_key = cast(ZoneInfo, fw_tz).key
            if match_key not in matches:
                matches.append(match_key)
    return matches


def _is_same_timezone(tzone1: tzinfo, tzone2: tzinfo) -> bool:
    """Return true if the timezones have the same utcffset and dst offset.

    Iot devices only support a limited static list of IANA timezones; this is used to
    check if a static timezone matches the same utc offset and dst settings.
    """
    now = datetime.now()
    start_day = datetime(now.year, 1, 1, 12)
    for i in range(365):
        the_day = start_day + timedelta(days=i)
        if tzone1.utcoffset(the_day) != tzone2.utcoffset(the_day):
            return False
    return True


TIMEZONE_INDEX = {
    0: "Etc/GMT+12",
    1: "Pacific/Samoa",
    2: "US/Hawaii",
    3: "US/Alaska",
    4: "Mexico/BajaNorte",
    5: "Etc/GMT+8",
    6: "PST8PDT",
    7: "US/Arizona",
    8: "America/Mazatlan",
    9: "MST",
    10: "MST7MDT",
    11: "Mexico/General",
    12: "Etc/GMT+6",
    13: "CST6CDT",
    14: "America/Monterrey",
    15: "Canada/Saskatchewan",
    16: "America/Bogota",
    17: "Etc/GMT+5",
    18: "EST",
    19: "America/Indiana/Indianapolis",
    20: "America/Caracas",
    21: "America/Asuncion",
    22: "Etc/GMT+4",
    23: "Canada/Atlantic",
    24: "America/Cuiaba",
    25: "Brazil/West",
    26: "America/Santiago",
    27: "Canada/Newfoundland",
    28: "America/Sao_Paulo",
    29: "America/Argentina/Buenos_Aires",
    30: "America/Cayenne",
    31: "America/Miquelon",
    32: "America/Montevideo",
    33: "Chile/Continental",
    34: "Etc/GMT+2",
    35: "Atlantic/Azores",
    36: "Atlantic/Cape_Verde",
    37: "Africa/Casablanca",
    38: "UCT",
    39: "GB",
    40: "Africa/Monrovia",
    41: "Europe/Amsterdam",
    42: "Europe/Belgrade",
    43: "Europe/Brussels",
    44: "Europe/Sarajevo",
    45: "Africa/Lagos",
    46: "Africa/Windhoek",
    47: "Asia/Amman",
    48: "Europe/Athens",
    49: "Asia/Beirut",
    50: "Africa/Cairo",
    51: "Asia/Damascus",
    52: "EET",
    53: "Africa/Harare",
    54: "Europe/Helsinki",
    55: "Asia/Istanbul",
    56: "Asia/Jerusalem",
    57: "Europe/Kaliningrad",
    58: "Africa/Tripoli",
    59: "Asia/Baghdad",
    60: "Asia/Kuwait",
    61: "Europe/Minsk",
    62: "Europe/Moscow",
    63: "Africa/Nairobi",
    64: "Asia/Tehran",
    65: "Asia/Muscat",
    66: "Asia/Baku",
    67: "Europe/Samara",
    68: "Indian/Mauritius",
    69: "Asia/Tbilisi",
    70: "Asia/Yerevan",
    71: "Asia/Kabul",
    72: "Asia/Ashgabat",
    73: "Asia/Yekaterinburg",
    74: "Asia/Karachi",
    75: "Asia/Kolkata",
    76: "Asia/Colombo",
    77: "Asia/Kathmandu",
    78: "Asia/Almaty",
    79: "Asia/Dhaka",
    80: "Asia/Novosibirsk",
    81: "Asia/Rangoon",
    82: "Asia/Bangkok",
    83: "Asia/Krasnoyarsk",
    84: "Asia/Chongqing",
    85: "Asia/Irkutsk",
    86: "Asia/Singapore",
    87: "Australia/Perth",
    88: "Asia/Taipei",
    89: "Asia/Ulaanbaatar",
    90: "Asia/Tokyo",
    91: "Asia/Seoul",
    92: "Asia/Yakutsk",
    93: "Australia/Adelaide",
    94: "Australia/Darwin",
    95: "Australia/Brisbane",
    96: "Australia/Canberra",
    97: "Pacific/Guam",
    98: "Australia/Hobart",
    99: "Antarctica/DumontDUrville",
    100: "Asia/Magadan",
    101: "Asia/Srednekolymsk",
    102: "Etc/GMT-11",
    103: "Asia/Anadyr",
    104: "Pacific/Auckland",
    105: "Etc/GMT-12",
    106: "Pacific/Fiji",
    107: "Etc/GMT-13",
    108: "Pacific/Apia",
    109: "Etc/GMT-14",
}



"""Modules for SMART devices."""

from ..effects import SmartLightEffect
from .alarm import Alarm
from .autooff import AutoOff
from .batterysensor import BatterySensor
from .brightness import Brightness
from .childdevice import ChildDevice
from .childlock import ChildLock
from .childprotection import ChildProtection
from .childsetup import ChildSetup
from .clean import Clean
from .cleanrecords import CleanRecords
from .cloud import Cloud
from .color import Color
from .colortemperature import ColorTemperature
from .consumables import Consumables
from .contactsensor import ContactSensor
from .devicemodule import DeviceModule
from .dustbin import Dustbin
from .energy import Energy
from .fan import Fan
from .firmware import Firmware
from .frostprotection import FrostProtection
from .homekit import HomeKit
from .humiditysensor import HumiditySensor
from .led import Led
from .light import Light
from .lighteffect import LightEffect
from .lightpreset import LightPreset
from .lightstripeffect import LightStripEffect
from .lighttransition import LightTransition
from .matter import Matter
from .mop import Mop
from .motionsensor import MotionSensor
from .overheatprotection import OverheatProtection
from .powerprotection import PowerProtection
from .reportmode import ReportMode
from .speaker import Speaker
from .temperaturecontrol import TemperatureControl
from .temperaturesensor import TemperatureSensor
from .thermostat import Thermostat
from .time import Time
from .triggerlogs import TriggerLogs
from .waterleaksensor import WaterleakSensor

__all__ = [
    "Alarm",
    "Time",
    "Energy",
    "DeviceModule",
    "ChildDevice",
    "ChildLock",
    "ChildSetup",
    "BatterySensor",
    "HumiditySensor",
    "TemperatureSensor",
    "TemperatureControl",
    "ChildProtection",
    "ReportMode",
    "AutoOff",
    "Led",
    "Brightness",
    "Fan",
    "LightPreset",
    "Firmware",
    "Cloud",
    "Light",
    "LightEffect",
    "LightStripEffect",
    "LightTransition",
    "ColorTemperature",
    "Color",
    "WaterleakSensor",
    "ContactSensor",
    "MotionSensor",
    "TriggerLogs",
    "FrostProtection",
    "Thermostat",
    "Clean",
    "Consumables",
    "CleanRecords",
    "SmartLightEffect",
    "PowerProtection",
    "OverheatProtection",
    "Speaker",
    "HomeKit",
    "Matter",
    "Dustbin",
    "Mop",
]



import base64
from unittest.mock import ANY

import aiohttp
import pytest
from yarl import URL

from kasa.credentials import DEFAULT_CREDENTIALS, Credentials, get_default_credentials
from kasa.deviceconfig import DeviceConfig
from kasa.exceptions import KasaException
from kasa.httpclient import HttpClient
from kasa.json import dumps as json_dumps
from kasa.transports.linkietransport import LinkieTransportV2

KASACAM_REQUEST_PLAINTEXT = '{"smartlife.cam.ipcamera.dateTime":{"get_status":{}}}'
KASACAM_RESPONSE_ENCRYPTED = "0PKG74LnnfKc+dvhw5bCgaycqZOjk7Gdv96syaiKsJLTvtupwKPC7aPGse632KrB48/tiPiX9JzDsNW2lK6fqZCgmKuZoZGh3A=="
KASACAM_RESPONSE_ERROR = '{"smartlife.cam.ipcamera.cloud": {"get_inf": {"err_code": -10008, "err_msg": "Unsupported API call."}}}'
KASA_DEFAULT_CREDENTIALS_HASH = "YWRtaW46MjEyMzJmMjk3YTU3YTVhNzQzODk0YTBlNGE4MDFmYzM="


async def test_working(mocker):
    """No errors with an expected request/response."""
    host = "127.0.0.1"
    mock_linkie_device = MockLinkieDevice(host)
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_linkie_device.post
    )
    transport_no_creds = LinkieTransportV2(config=DeviceConfig(host))

    response = await transport_no_creds.send(KASACAM_REQUEST_PLAINTEXT)
    assert response == {
        "timezone": "UTC-05:00",
        "area": "America/New_York",
        "epoch_sec": 1690832800,
    }


async def test_credentials_hash(mocker):
    """Ensure the default credentials are always passed as Basic Auth."""
    # Test without credentials input

    host = "127.0.0.1"
    mock_linkie_device = MockLinkieDevice(host)
    mock_post = mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_linkie_device.post
    )
    transport_no_creds = LinkieTransportV2(config=DeviceConfig(host))
    await transport_no_creds.send(KASACAM_REQUEST_PLAINTEXT)
    mock_post.assert_called_once_with(
        URL(f"https://{host}:10443/data/LINKIE2.json"),
        params=None,
        data=ANY,
        json=None,
        timeout=ANY,
        cookies=None,
        headers={
            "Authorization": "Basic " + _generate_kascam_basic_auth(),
            "Content-Type": "application/x-www-form-urlencoded",
        },
        ssl=ANY,
    )

    assert transport_no_creds.credentials_hash == KASA_DEFAULT_CREDENTIALS_HASH
    # Test with credentials input

    transport_with_creds = LinkieTransportV2(
        config=DeviceConfig(host, credentials=Credentials("Admin", "password"))
    )
    mock_post.reset_mock()

    await transport_with_creds.send(KASACAM_REQUEST_PLAINTEXT)
    mock_post.assert_called_once_with(
        URL(f"https://{host}:10443/data/LINKIE2.json"),
        params=None,
        data=ANY,
        json=None,
        timeout=ANY,
        cookies=None,
        headers={
            "Authorization": "Basic " + _generate_kascam_basic_auth(),
            "Content-Type": "application/x-www-form-urlencoded",
        },
        ssl=ANY,
    )


@pytest.mark.parametrize(
    ("return_status", "return_data", "expected"),
    [
        (500, KASACAM_RESPONSE_ENCRYPTED, "500"),
        (200, "AAAAAAAAAAAAAAAAAAAAAAAA", "Unable to read response"),
        (200, KASACAM_RESPONSE_ERROR, "Unsupported API call"),
    ],
)
async def test_exceptions(mocker, return_status, return_data, expected):
    """Test a variety of possible responses from the device."""
    host = "127.0.0.1"
    transport = LinkieTransportV2(config=DeviceConfig(host))
    mock_linkie_device = MockLinkieDevice(
        host, status_code=return_status, response=return_data
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_linkie_device.post
    )

    with pytest.raises(KasaException, match=expected):
        await transport.send(KASACAM_REQUEST_PLAINTEXT)


def _generate_kascam_basic_auth():
    creds = get_default_credentials(DEFAULT_CREDENTIALS["KASACAMERA"])
    creds_combined = f"{creds.username}:{creds.password}"
    return base64.b64encode(creds_combined.encode()).decode()


class MockLinkieDevice:
    """Based on MockSslDevice."""

    class _mock_response:
        def __init__(self, status, request: dict):
            self.status = status
            self._json = request

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_t, exc_v, exc_tb):
            pass

        async def read(self):
            if isinstance(self._json, dict):
                return json_dumps(self._json).encode()
            return self._json

    def __init__(self, host, *, status_code=200, response=KASACAM_RESPONSE_ENCRYPTED):
        self.host = host
        self.http_client = HttpClient(DeviceConfig(self.host))
        self.status_code = status_code
        self.response = response

    async def post(
        self, url: URL, *, headers=None, params=None, json=None, data=None, **__
    ):
        return self._mock_response(self.status_code, self.response)



"""Tests for smartcam pet detection module."""

from __future__ import annotations

from kasa import Device
from kasa.smartcam.smartcammodule import SmartCamModule

from ...device_fixtures import parametrize

petdetection = parametrize(
    "has pet detection",
    component_filter="petDetection",
    protocol_filter={"SMARTCAM"},
)


@petdetection
async def test_petdetection(dev: Device):
    """Test device pet detection."""
    pet = dev.modules.get(SmartCamModule.SmartCamPetDetection)
    assert pet

    pde_feat = dev.features.get("pet_detection")
    assert pde_feat

    original_enabled = pet.enabled

    try:
        await pet.set_enabled(not original_enabled)
        await dev.update()
        assert pet.enabled is not original_enabled
        assert pde_feat.value is not original_enabled

        await pet.set_enabled(original_enabled)
        await dev.update()
        assert pet.enabled is original_enabled
        assert pde_feat.value is original_enabled

        await pde_feat.set_value(not original_enabled)
        await dev.update()
        assert pet.enabled is not original_enabled
        assert pde_feat.value is not original_enabled

    finally:
        await pet.set_enabled(original_enabled)



import pytest

from kasa import DeviceType, Module
from kasa.iot import IotLightStrip
from kasa.iot.modules import LightEffect
from tests.conftest import lightstrip_iot


@lightstrip_iot
async def test_lightstrip_length(dev: IotLightStrip):
    assert dev.device_type == DeviceType.LightStrip
    assert dev.length == dev.sys_info["length"]


@lightstrip_iot
async def test_lightstrip_effect(dev: IotLightStrip):
    le: LightEffect = dev.modules[Module.LightEffect]
    assert isinstance(le._deprecated_effect, dict)
    for k in ["brightness", "custom", "enable", "id", "name"]:
        assert k in le._deprecated_effect


@lightstrip_iot
async def test_effects_lightstrip_set_effect(dev: IotLightStrip):
    le: LightEffect = dev.modules[Module.LightEffect]
    with pytest.raises(
        ValueError, match="The effect Not real is not a built in effect"
    ):
        await le.set_effect("Not real")

    await le.set_effect("Candy Cane")
    await dev.update()
    assert le.effect == "Candy Cane"


@lightstrip_iot
@pytest.mark.parametrize("brightness", [100, 50])
async def test_effects_lightstrip_set_effect_brightness(
    dev: IotLightStrip, brightness, mocker
):
    query_helper = mocker.patch("kasa.iot.IotLightStrip._query_helper")
    le: LightEffect = dev.modules[Module.LightEffect]

    # test that default brightness works (100 for candy cane)
    if brightness == 100:
        await le.set_effect("Candy Cane")
    else:
        await le.set_effect("Candy Cane", brightness=brightness)

    args, kwargs = query_helper.call_args_list[0]
    payload = args[2]
    assert payload["brightness"] == brightness


@lightstrip_iot
@pytest.mark.parametrize("transition", [500, 1000])
async def test_effects_lightstrip_set_effect_transition(
    dev: IotLightStrip, transition, mocker
):
    query_helper = mocker.patch("kasa.iot.IotLightStrip._query_helper")
    le: LightEffect = dev.modules[Module.LightEffect]

    # test that default (500 for candy cane) transition works
    if transition == 500:
        await le.set_effect("Candy Cane")
    else:
        await le.set_effect("Candy Cane", transition=transition)

    args, kwargs = query_helper.call_args_list[0]
    payload = args[2]
    assert payload["transition"] == transition


@lightstrip_iot
async def test_effects_lightstrip_has_effects(dev: IotLightStrip):
    le: LightEffect = dev.modules[Module.LightEffect]
    assert le is not None
    assert le.effect_list


@lightstrip_iot
def test_device_type_lightstrip(dev):
    assert dev.device_type == DeviceType.LightStrip



import pytest

from kasa.smart.modules import TemperatureSensor

from ...device_fixtures import parametrize

temperature = parametrize(
    "has temperature", component_filter="temperature", protocol_filter={"SMART.CHILD"}
)

temperature_warning = parametrize(
    "has temperature warning",
    component_filter="comfort_temperature",
    protocol_filter={"SMART.CHILD"},
)


@temperature
@pytest.mark.parametrize(
    ("feature", "type"),
    [
        ("temperature", float),
        ("temperature_unit", str),
    ],
)
async def test_temperature_features(dev, feature, type):
    """Test that features are registered and work as expected."""
    temp_module: TemperatureSensor = dev.modules["TemperatureSensor"]

    prop = getattr(temp_module, feature)
    assert isinstance(prop, type)

    feat = dev.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)


@temperature_warning
async def test_temperature_warning(dev):
    """Test that features are registered and work as expected."""
    temp_module: TemperatureSensor = dev.modules["TemperatureSensor"]

    assert hasattr(temp_module, "temperature_warning")
    assert isinstance(temp_module.temperature_warning, bool)

    feat = dev.features["temperature_warning"]
    assert feat.value == temp_module.temperature_warning
    assert isinstance(feat.value, bool)



"""Implementation of vacuum mop."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import Annotated

from ...feature import Feature
from ...module import FeatureAttribute
from ..smartmodule import SmartModule

_LOGGER = logging.getLogger(__name__)


class Waterlevel(IntEnum):
    """Water level for mopping."""

    Disable = 0
    Low = 1
    Medium = 2
    High = 3


class Mop(SmartModule):
    """Implementation of vacuum mop."""

    REQUIRED_COMPONENT = "mop"

    def _initialize_features(self) -> None:
        """Initialize features."""
        self._add_feature(
            Feature(
                self._device,
                id="mop_attached",
                name="Mop attached",
                container=self,
                icon="mdi:square-rounded",
                attribute_getter="mop_attached",
                category=Feature.Category.Info,
                type=Feature.BinarySensor,
            )
        )

        self._add_feature(
            Feature(
                self._device,
                id="mop_waterlevel",
                name="Mop water level",
                container=self,
                attribute_getter="waterlevel",
                attribute_setter="set_waterlevel",
                icon="mdi:water",
                choices_getter=lambda: list(Waterlevel.__members__),
                category=Feature.Category.Config,
                type=Feature.Type.Choice,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {
            "getMopState": {},
            "getCleanAttr": {"type": "global"},
        }

    @property
    def mop_attached(self) -> bool:
        """Return True if mop is attached."""
        return self.data["getMopState"]["mop_state"]

    @property
    def _settings(self) -> dict:
        """Return settings settings."""
        return self.data["getCleanAttr"]

    @property
    def waterlevel(self) -> Annotated[str, FeatureAttribute()]:
        """Return water level."""
        return Waterlevel(int(self._settings["cistern"])).name

    async def set_waterlevel(self, mode: str) -> Annotated[dict, FeatureAttribute()]:
        """Set waterlevel mode."""
        name_to_value = {x.name: x.value for x in Waterlevel}
        if mode not in name_to_value:
            raise ValueError("Invalid waterlevel %s, available %s", mode, name_to_value)

        settings = self._settings.copy()
        settings["cistern"] = name_to_value[mode]
        return await self.call("setCleanAttr", settings)



from __future__ import annotations

import pytest
from pytest_mock import MockerFixture

from kasa import Module
from kasa.smart import SmartDevice
from kasa.smart.modules.mop import Waterlevel

from ...device_fixtures import get_parent_and_child_modules, parametrize

mop = parametrize("has mop", component_filter="mop", protocol_filter={"SMART"})


@mop
@pytest.mark.parametrize(
    ("feature", "prop_name", "type"),
    [
        ("mop_attached", "mop_attached", bool),
        ("mop_waterlevel", "waterlevel", str),
    ],
)
async def test_features(dev: SmartDevice, feature: str, prop_name: str, type: type):
    """Test that features are registered and work as expected."""
    mod = next(get_parent_and_child_modules(dev, Module.Mop))
    assert mod is not None

    prop = getattr(mod, prop_name)
    assert isinstance(prop, type)

    feat = mod._device.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)


@mop
async def test_mop_waterlevel(dev: SmartDevice, mocker: MockerFixture):
    """Test dust mode."""
    mop_module = next(get_parent_and_child_modules(dev, Module.Mop))
    call = mocker.spy(mop_module, "call")

    waterlevel = mop_module._device.features["mop_waterlevel"]
    assert mop_module.waterlevel == waterlevel.value

    new_level = Waterlevel.High
    await mop_module.set_waterlevel(new_level.name)

    params = mop_module._settings.copy()
    params["cistern"] = new_level.value

    call.assert_called_with("setCleanAttr", params)

    await dev.update()

    assert mop_module.waterlevel == new_level.name

    with pytest.raises(ValueError, match="Invalid waterlevel"):
        await mop_module.set_waterlevel("invalid")



"""Interact with modules.

Modules are implemented by devices to encapsulate sets of functionality like
Light, AutoOff, Firmware etc.

>>> from kasa import Discover, Module
>>>
>>> dev = await Discover.discover_single(
>>>     "127.0.0.3",
>>>     username="user@example.com",
>>>     password="great_password"
>>> )
>>> await dev.update()
>>> print(dev.alias)
Living Room Bulb

To see whether a device supports a group of functionality
check for the existence of the module:

>>> if light := dev.modules.get("Light"):
>>>     print(light.brightness)
100

.. include:: ../featureattributes.md
   :parser: myst_parser.sphinx_

To see whether a device supports specific functionality, you can check whether the
module has that feature:

>>> if light.has_feature("hsv"):
>>>     print(light.hsv)
HSV(hue=0, saturation=100, value=100)

If you know or expect the module to exist you can access by index:

>>> light_preset = dev.modules["LightPreset"]
>>> print(light_preset.preset_list)
['Not set', 'Light preset 1', 'Light preset 2', 'Light preset 3',\
 'Light preset 4', 'Light preset 5', 'Light preset 6', 'Light preset 7']

Modules support typing via the Module names in Module:

>>> from typing import reveal_type, TYPE_CHECKING
>>> light_effect = dev.modules.get("LightEffect")
>>> light_effect_typed = dev.modules.get(Module.LightEffect)
>>> if TYPE_CHECKING:
>>>     reveal_type(light_effect)  # Static checker will reveal: str
>>>     reveal_type(light_effect_typed)  # Static checker will reveal: LightEffect

"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import cache
from typing import (
    TYPE_CHECKING,
    Final,
    TypeVar,
    get_type_hints,
)

from .exceptions import KasaException
from .feature import Feature
from .modulemapping import ModuleName

if TYPE_CHECKING:
    from . import interfaces
    from .device import Device
    from .iot import modules as iot
    from .smart import modules as smart
    from .smartcam import modules as smartcam

_LOGGER = logging.getLogger(__name__)

ModuleT = TypeVar("ModuleT", bound="Module")


class FeatureAttribute:
    """Class for annotating attributes bound to feature."""

    def __init__(self, feature_name: str | None = None) -> None:
        self.feature_name = feature_name

    def __repr__(self) -> str:
        return "FeatureAttribute"


class Module(ABC):
    """Base class implemention for all modules.

    The base classes should implement `query` to return the query they want to be
    executed during the regular update cycle.
    """

    # Common Modules
    Alarm: Final[ModuleName[interfaces.Alarm]] = ModuleName("Alarm")
    ChildSetup: Final[ModuleName[interfaces.ChildSetup]] = ModuleName("ChildSetup")
    Energy: Final[ModuleName[interfaces.Energy]] = ModuleName("Energy")
    Fan: Final[ModuleName[interfaces.Fan]] = ModuleName("Fan")
    LightEffect: Final[ModuleName[interfaces.LightEffect]] = ModuleName("LightEffect")
    Led: Final[ModuleName[interfaces.Led]] = ModuleName("Led")
    Light: Final[ModuleName[interfaces.Light]] = ModuleName("Light")
    LightPreset: Final[ModuleName[interfaces.LightPreset]] = ModuleName("LightPreset")
    Thermostat: Final[ModuleName[interfaces.Thermostat]] = ModuleName("Thermostat")
    Time: Final[ModuleName[interfaces.Time]] = ModuleName("Time")

    # IOT only Modules
    IotAmbientLight: Final[ModuleName[iot.AmbientLight]] = ModuleName("ambient")
    IotAntitheft: Final[ModuleName[iot.Antitheft]] = ModuleName("anti_theft")
    IotCountdown: Final[ModuleName[iot.Countdown]] = ModuleName("countdown")
    IotDimmer: Final[ModuleName[iot.Dimmer]] = ModuleName("dimmer")
    IotMotion: Final[ModuleName[iot.Motion]] = ModuleName("motion")
    IotSchedule: Final[ModuleName[iot.Schedule]] = ModuleName("schedule")
    IotUsage: Final[ModuleName[iot.Usage]] = ModuleName("usage")
    IotCloud: Final[ModuleName[iot.Cloud]] = ModuleName("cloud")

    # SMART only Modules
    AutoOff: Final[ModuleName[smart.AutoOff]] = ModuleName("AutoOff")
    BatterySensor: Final[ModuleName[smart.BatterySensor]] = ModuleName("BatterySensor")
    Brightness: Final[ModuleName[smart.Brightness]] = ModuleName("Brightness")
    ChildDevice: Final[ModuleName[smart.ChildDevice]] = ModuleName("ChildDevice")
    Cloud: Final[ModuleName[smart.Cloud]] = ModuleName("Cloud")
    Color: Final[ModuleName[smart.Color]] = ModuleName("Color")
    ColorTemperature: Final[ModuleName[smart.ColorTemperature]] = ModuleName(
        "ColorTemperature"
    )
    ContactSensor: Final[ModuleName[smart.ContactSensor]] = ModuleName("ContactSensor")
    DeviceModule: Final[ModuleName[smart.DeviceModule]] = ModuleName("DeviceModule")
    Firmware: Final[ModuleName[smart.Firmware]] = ModuleName("Firmware")
    FrostProtection: Final[ModuleName[smart.FrostProtection]] = ModuleName(
        "FrostProtection"
    )
    HumiditySensor: Final[ModuleName[smart.HumiditySensor]] = ModuleName(
        "HumiditySensor"
    )
    LightTransition: Final[ModuleName[smart.LightTransition]] = ModuleName(
        "LightTransition"
    )
    MotionSensor: Final[ModuleName[smart.MotionSensor]] = ModuleName("MotionSensor")
    ReportMode: Final[ModuleName[smart.ReportMode]] = ModuleName("ReportMode")
    SmartLightEffect: Final[ModuleName[smart.SmartLightEffect]] = ModuleName(
        "LightEffect"
    )
    IotLightEffect: Final[ModuleName[iot.LightEffect]] = ModuleName("LightEffect")
    TemperatureSensor: Final[ModuleName[smart.TemperatureSensor]] = ModuleName(
        "TemperatureSensor"
    )
    TemperatureControl: Final[ModuleName[smart.TemperatureControl]] = ModuleName(
        "TemperatureControl"
    )
    WaterleakSensor: Final[ModuleName[smart.WaterleakSensor]] = ModuleName(
        "WaterleakSensor"
    )
    ChildProtection: Final[ModuleName[smart.ChildProtection]] = ModuleName(
        "ChildProtection"
    )
    ChildLock: Final[ModuleName[smart.ChildLock]] = ModuleName("ChildLock")
    TriggerLogs: Final[ModuleName[smart.TriggerLogs]] = ModuleName("TriggerLogs")
    PowerProtection: Final[ModuleName[smart.PowerProtection]] = ModuleName(
        "PowerProtection"
    )

    HomeKit: Final[ModuleName[smart.HomeKit]] = ModuleName("HomeKit")
    Matter: Final[ModuleName[smart.Matter]] = ModuleName("Matter")

    # SMARTCAM only modules
    Camera: Final[ModuleName[smartcam.Camera]] = ModuleName("Camera")
    LensMask: Final[ModuleName[smartcam.LensMask]] = ModuleName("LensMask")

    # Vacuum modules
    Clean: Final[ModuleName[smart.Clean]] = ModuleName("Clean")
    Consumables: Final[ModuleName[smart.Consumables]] = ModuleName("Consumables")
    Dustbin: Final[ModuleName[smart.Dustbin]] = ModuleName("Dustbin")
    Speaker: Final[ModuleName[smart.Speaker]] = ModuleName("Speaker")
    Mop: Final[ModuleName[smart.Mop]] = ModuleName("Mop")
    CleanRecords: Final[ModuleName[smart.CleanRecords]] = ModuleName("CleanRecords")

    def __init__(self, device: Device, module: str) -> None:
        self._device = device
        self._module = module
        self._module_features: dict[str, Feature] = {}

    @property
    def device(self) -> Device:
        """Return the device exposing the module."""
        return self._device

    @property
    def _all_features(self) -> dict[str, Feature]:
        """Get the features for this module and any sub modules."""
        return self._module_features

    def has_feature(self, attribute: str | property | Callable) -> bool:
        """Return True if the module attribute feature is supported."""
        return bool(self.get_feature(attribute))

    def get_feature(self, attribute: str | property | Callable) -> Feature | None:
        """Get Feature for a module attribute or None if not supported."""
        return _get_bound_feature(self, attribute)

    @abstractmethod
    def query(self) -> dict:
        """Query to execute during the update cycle.

        The inheriting modules implement this to include their wanted
        queries to the query that gets executed when Device.update() gets called.
        """

    @property
    @abstractmethod
    def data(self) -> dict:
        """Return the module specific raw data from the last update."""

    def _initialize_features(self) -> None:  # noqa: B027
        """Initialize features after the initial update.

        This can be implemented if features depend on module query responses.
        It will only be called once per module and will always be called
        after *_post_update_hook* has been called for every device module and its
        children's modules.
        """

    async def _post_update_hook(self) -> None:  # noqa: B027
        """Perform actions after a device update.

        This can be implemented if a module needs to perform actions each time
        the device has updated like generating collections for property access.
        It will be called after every update and will be called prior to
        *_initialize_features* on the first update.
        """

    def _add_feature(self, feature: Feature) -> None:
        """Add module feature."""
        id_ = feature.id
        if id_ in self._module_features:
            raise KasaException(f"Duplicate id detected {id_}")
        self._module_features[id_] = feature

    def __repr__(self) -> str:
        return (
            f"<Module {self.__class__.__name__} ({self._module})"
            f" for {self._device.host}>"
        )


def _get_feature_attribute(attribute: property | Callable) -> FeatureAttribute | None:
    """Check if an attribute is bound to a feature with FeatureAttribute."""
    if isinstance(attribute, property):
        hints = get_type_hints(attribute.fget, include_extras=True)
    else:
        hints = get_type_hints(attribute, include_extras=True)

    if (return_hints := hints.get("return")) and hasattr(return_hints, "__metadata__"):
        metadata = hints["return"].__metadata__
        for meta in metadata:
            if isinstance(meta, FeatureAttribute):
                return meta

    return None


@cache
def _get_bound_feature(
    module: Module, attribute: str | property | Callable
) -> Feature | None:
    """Get Feature for a bound property or None if not supported."""
    if not isinstance(attribute, str):
        if isinstance(attribute, property):
            # Properties have __name__ in 3.13 so this could be simplified
            # when only 3.13 supported
            attribute_name = attribute.fget.__name__  # type: ignore[union-attr]
        else:
            attribute_name = attribute.__name__
        attribute_callable = attribute
    else:
        if TYPE_CHECKING:
            assert isinstance(attribute, str)
        attribute_name = attribute
        attribute_callable = getattr(module.__class__, attribute, None)  # type: ignore[assignment]
        if not attribute_callable:
            raise KasaException(
                f"No attribute named {attribute_name} in "
                f"module {module.__class__.__name__}"
            )

    if not (fa := _get_feature_attribute(attribute_callable)):
        raise KasaException(
            f"Attribute {attribute_name} of module {module.__class__.__name__}"
            " is not bound to a feature"
        )

    # If a feature_name was passed to the FeatureAttribute use that to check
    # for the feature. Otherwise check the getters and setters in the features
    if fa.feature_name:
        return module._all_features.get(fa.feature_name)

    check = {attribute_name, attribute_callable}
    for feature in module._all_features.values():
        if (getter := feature.attribute_getter) and getter in check:
            return feature

        if (setter := feature.attribute_setter) and setter in check:
            return feature

    return None



import pytest
from pytest_mock import MockerFixture

from kasa import Module, SmartDevice

from ...device_fixtures import get_parent_and_child_modules, parametrize

powerprotection = parametrize(
    "has powerprotection",
    component_filter="power_protection",
    protocol_filter={"SMART"},
)


@powerprotection
@pytest.mark.parametrize(
    ("feature", "prop_name", "type"),
    [
        ("overloaded", "overloaded", bool),
        ("power_protection_threshold", "protection_threshold", int),
    ],
)
async def test_features(dev, feature, prop_name, type):
    """Test that features are registered and work as expected."""
    powerprot = next(get_parent_and_child_modules(dev, Module.PowerProtection))
    assert powerprot
    device = powerprot._device

    prop = getattr(powerprot, prop_name)
    assert isinstance(prop, type)

    feat = device.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)


@powerprotection
async def test_set_enable(dev: SmartDevice, mocker: MockerFixture):
    """Test enable."""
    powerprot = next(get_parent_and_child_modules(dev, Module.PowerProtection))
    assert powerprot
    device = powerprot._device

    original_enabled = powerprot.enabled
    original_threshold = powerprot.protection_threshold

    try:
        # Simple enable with an existing threshold
        call_spy = mocker.spy(powerprot, "call")
        await powerprot.set_enabled(True)
        params = {
            "enabled": True,
            "protection_power": mocker.ANY,
        }
        call_spy.assert_called_with("set_protection_power", params)

        # Enable with no threshold param when 0
        call_spy.reset_mock()
        await powerprot.set_protection_threshold(0)
        await device.update()
        await powerprot.set_enabled(True)
        params = {
            "enabled": True,
            "protection_power": int(powerprot._max_power / 2),
        }
        call_spy.assert_called_with("set_protection_power", params)

        # Enable false should not update the threshold
        call_spy.reset_mock()
        await powerprot.set_protection_threshold(0)
        await device.update()
        await powerprot.set_enabled(False)
        params = {
            "enabled": False,
            "protection_power": 0,
        }
        call_spy.assert_called_with("set_protection_power", params)

    finally:
        await powerprot.set_enabled(original_enabled, threshold=original_threshold)


@powerprotection
async def test_set_threshold(dev: SmartDevice, mocker: MockerFixture):
    """Test enable."""
    powerprot = next(get_parent_and_child_modules(dev, Module.PowerProtection))
    assert powerprot

    call_spy = mocker.spy(powerprot, "call")
    await powerprot.set_protection_threshold(123)
    params = {
        "enabled": mocker.ANY,
        "protection_power": 123,
    }
    call_spy.assert_called_with("set_protection_power", params)

    with pytest.raises(ValueError, match="Threshold out of range"):
        await powerprot.set_protection_threshold(-10)



"""Script for testing update performance on devices."""

import asyncio
import time

import asyncclick as click
import pandas as pd

from kasa import Discover


async def _update(dev, lock=None):
    if lock is not None:
        await lock.acquire()
        await asyncio.sleep(2)
    try:
        start_time = time.time()
        # print("%s >> Updating" % id(dev))
        await dev.update()
        # print("%s >> done in %s" % (id(dev), time.time() - start_time))
        return {"id": f"{id(dev)}-{dev.model}", "took": (time.time() - start_time)}
    finally:
        if lock is not None:
            lock.release()


async def _update_concurrently(devs):
    start_time = time.time()
    update_futures = [asyncio.ensure_future(_update(dev)) for dev in devs]
    await asyncio.gather(*update_futures)
    return {"type": "concurrently", "took": (time.time() - start_time)}


async def _update_sequentially(devs):
    start_time = time.time()

    for dev in devs:
        await _update(dev)

    return {"type": "sequential", "took": (time.time() - start_time)}


@click.command()
@click.argument("addrs", nargs=-1)
@click.option("--rounds", default=5)
async def main(addrs, rounds):
    """Test update performance on given devices."""
    print(f"Running {rounds} rounds on {addrs}")
    devs = []

    for addr in addrs:
        try:
            dev = await Discover.discover_single(addr)
            devs.append(dev)
        except Exception as ex:
            print(f"unable to add {addr}: {ex}")

    data = []
    test_gathered = True

    if test_gathered:
        print("=== Testing using gather on all devices ===")
        for _i in range(rounds):
            data.append(await _update_concurrently(devs))
            await asyncio.sleep(2)

        await asyncio.sleep(5)

        for _i in range(rounds):
            data.append(await _update_sequentially(devs))
            await asyncio.sleep(2)

        df = pd.DataFrame(data)
        print(df.groupby("type").describe())

    print("=== Testing per-device performance ===")

    futs = []
    data = []
    locks = {dev: asyncio.Lock() for dev in devs}
    for _i in range(rounds):
        for dev in devs:
            futs.append(asyncio.ensure_future(_update(dev, locks[dev])))

    for fut in asyncio.as_completed(futs):
        res = await fut
        data.append(res)

    df = pd.DataFrame(data)
    print(df.groupby("id").describe())


if __name__ == "__main__":
    main(_anyio_backend="asyncio")



"""Module for led controls."""

from __future__ import annotations

from ...interfaces.led import Led as LedInterface
from ..smartmodule import SmartModule, allow_update_after


class Led(SmartModule, LedInterface):
    """Implementation of led controls."""

    REQUIRED_COMPONENT = "led"
    QUERY_GETTER_NAME = "get_led_info"
    # Led queries can cause device to crash on P100
    MINIMUM_UPDATE_INTERVAL_SECS = 60 * 60

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {self.QUERY_GETTER_NAME: None}

    @property
    def mode(self) -> str:
        """LED mode setting.

        "always", "never", "night_mode"
        """
        return self.data["led_rule"]

    @property
    def led(self) -> bool:
        """Return current led status."""
        return self.data["led_rule"] != "never"

    @allow_update_after
    async def set_led(self, enable: bool) -> dict:
        """Set led.

        This should probably be a select with always/never/nightmode.
        """
        rule = "always" if enable else "never"
        return await self.call("set_led_info", dict(self.data, **{"led_rule": rule}))

    @property
    def night_mode_settings(self) -> dict:
        """Night mode settings."""
        return {
            "start": self.data["start_time"],
            "end": self.data["end_time"],
            "type": self.data["night_mode_type"],
            "sunrise_offset": self.data["sunrise_offset"],
            "sunset_offset": self.data["sunset_offset"],
        }



"""Module for dump_devinfo tests."""

import copy

import pytest

from devtools.dump_devinfo import (
    _wrap_redactors,
    get_legacy_fixture,
    get_smart_fixtures,
)
from kasa.iot import IotDevice
from kasa.protocols import IotProtocol
from kasa.protocols.protocol import redact_data
from kasa.protocols.smartprotocol import REDACTORS as SMART_REDACTORS
from kasa.smart import SmartDevice
from kasa.smartcam import SmartCamDevice
from kasa.smartcam.smartcamchild import CHILD_INFO_FROM_PARENT

from .conftest import (
    FixtureInfo,
    get_device_for_fixture,
    get_fixture_info,
    parametrize,
)

smart_fixtures = parametrize(
    "smart fixtures", protocol_filter={"SMART"}, fixture_name="fixture_info"
)
smartcam_fixtures = parametrize(
    "smartcam fixtures", protocol_filter={"SMARTCAM"}, fixture_name="fixture_info"
)
iot_fixtures = parametrize(
    "iot fixtures", protocol_filter={"IOT"}, fixture_name="fixture_info"
)


async def test_fixture_names(fixture_info: FixtureInfo):
    """Test that device info gets the right fixture names."""
    if fixture_info.protocol in {"SMARTCAM"}:
        device_info = SmartCamDevice._get_device_info(
            fixture_info.data,
            fixture_info.data.get("discovery_result", {}).get("result"),
        )
    elif fixture_info.protocol in {"SMART"}:
        device_info = SmartDevice._get_device_info(
            fixture_info.data,
            fixture_info.data.get("discovery_result", {}).get("result"),
        )
    elif fixture_info.protocol in {"SMART.CHILD"}:
        device_info = SmartDevice._get_device_info(fixture_info.data, None)
    else:
        device_info = IotDevice._get_device_info(fixture_info.data, None)

    region = f"({device_info.region})" if device_info.region else ""
    expected = f"{device_info.long_name}{region}_{device_info.hardware_version}_{device_info.firmware_version}.json"
    assert fixture_info.name == expected


@smart_fixtures
async def test_smart_fixtures(fixture_info: FixtureInfo):
    """Test that smart fixtures are created the same."""
    dev = await get_device_for_fixture(fixture_info, verbatim=True)
    assert isinstance(dev, SmartDevice)
    if dev.children:
        pytest.skip("Test not currently implemented for devices with children.")
    fixtures = await get_smart_fixtures(
        dev.protocol,
        discovery_info=fixture_info.data.get("discovery_result"),
        batch_size=5,
    )
    fixture_result = fixtures[0]

    assert fixture_info.data == fixture_result.data


def _normalize_child_device_ids(info: dict):
    """Scrubbed child device ids in hubs may not match ids in child fixtures.

    Different hub fixtures could create the same child fixture so we scrub
    them again for the purpose of the test.
    """
    if dev_info := info.get("get_device_info"):
        dev_info["device_id"] = "SCRUBBED"
    elif (
        dev_info := info.get("getDeviceInfo", {})
        .get("device_info", {})
        .get("basic_info")
    ):
        dev_info["dev_id"] = "SCRUBBED"


@smartcam_fixtures
async def test_smartcam_fixtures(fixture_info: FixtureInfo):
    """Test that smartcam fixtures are created the same."""
    dev = await get_device_for_fixture(fixture_info, verbatim=True)
    assert isinstance(dev, SmartCamDevice)

    created_fixtures = await get_smart_fixtures(
        dev.protocol,
        discovery_info=fixture_info.data.get("discovery_result"),
        batch_size=5,
    )
    fixture_result = created_fixtures.pop(0)

    assert fixture_info.data == fixture_result.data

    for created_child_fixture in created_fixtures:
        child_fixture_info = get_fixture_info(
            created_child_fixture.filename + ".json",
            created_child_fixture.protocol_suffix,
        )

        assert child_fixture_info

        _normalize_child_device_ids(created_child_fixture.data)

        saved_fixture_data = copy.deepcopy(child_fixture_info.data)
        _normalize_child_device_ids(saved_fixture_data)
        saved_fixture_data = {
            key: val for key, val in saved_fixture_data.items() if val != -1001
        }

        # Remove the child info from parent from the comparison because the
        # child may have been created by a different parent fixture
        saved_fixture_data.pop(CHILD_INFO_FROM_PARENT, None)
        created_cifp = created_child_fixture.data.pop(CHILD_INFO_FROM_PARENT, None)

        # Still check that the created child info from parent was redacted.
        # only smartcam children generate child_info_from_parent
        if created_cifp:
            redacted_cifp = redact_data(created_cifp, _wrap_redactors(SMART_REDACTORS))
            assert created_cifp == redacted_cifp

        assert saved_fixture_data == created_child_fixture.data


@iot_fixtures
async def test_iot_fixtures(fixture_info: FixtureInfo):
    """Test that iot fixtures are created the same."""
    # Iot fixtures often do not have enough data to perform a device update()
    # without missing info being added to suppress the update
    dev = await get_device_for_fixture(
        fixture_info, verbatim=True, update_after_init=False
    )
    assert isinstance(dev.protocol, IotProtocol)

    fixture = await get_legacy_fixture(
        dev.protocol, discovery_info=fixture_info.data.get("discovery_result")
    )
    fixture_result = fixture

    created_fixture = {
        key: val for key, val in fixture_result.data.items() if "err_code" not in val
    }
    saved_fixture = {
        key: val for key, val in fixture_info.data.items() if "err_code" not in val
    }
    assert saved_fixture == created_fixture



"""Implementation of report module."""

from __future__ import annotations

from ...feature import Feature
from ..smartmodule import SmartModule


class ReportMode(SmartModule):
    """Implementation of report module."""

    REQUIRED_COMPONENT = "report_mode"
    QUERY_GETTER_NAME = "get_report_mode"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="report_interval",
                name="Report interval",
                container=self,
                attribute_getter="report_interval",
                unit_getter=lambda: "s",
                category=Feature.Category.Debug,
                type=Feature.Type.Sensor,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {}

    @property
    def report_interval(self) -> int:
        """Reporting interval of a sensor device."""
        return self._device.sys_info["report_interval"]



"""Interact with feature.

Features are implemented by devices to represent individual pieces of functionality like
state, time, firmware.

>>> from kasa import Discover, Module
>>>
>>> dev = await Discover.discover_single(
>>>     "127.0.0.3",
>>>     username="user@example.com",
>>>     password="great_password"
>>> )
>>> await dev.update()
>>> print(dev.alias)
Living Room Bulb

Features allow for instrospection and can be interacted with as new features are added
to the API:

>>> for feature_id, feature in dev.features.items():
>>>     print(f"{feature.name} ({feature_id}): {feature.value}")
Device ID (device_id): 0000000000000000000000000000000000000000
State (state): True
Signal Level (signal_level): 2
RSSI (rssi): -52
SSID (ssid): #MASKED_SSID#
Reboot (reboot): <Action>
Device time (device_time): 2024-02-23 02:40:15+01:00
Brightness (brightness): 100
Cloud connection (cloud_connection): True
HSV (hsv): HSV(hue=0, saturation=100, value=100)
Color temperature (color_temperature): 2700
Auto update enabled (auto_update_enabled): False
Update available (update_available): None
Current firmware version (current_firmware_version): 1.1.6 Build 240130 Rel.173828
Available firmware version (available_firmware_version): None
Check latest firmware (check_latest_firmware): <Action>
Light effect (light_effect): Off
Light preset (light_preset): Not set
Smooth transition on (smooth_transition_on): 2
Smooth transition off (smooth_transition_off): 2
Overheated (overheated): False

To see whether a device supports a feature, check for the existence of it:

>>> if feature := dev.features.get("brightness"):
>>>     print(feature.value)
100

You can update the value of a feature

>>> await feature.set_value(50)
>>> await dev.update()
>>> print(feature.value)
50

Features have types that can be used for introspection:

>>> feature = dev.features["light_preset"]
>>> print(feature.type)
Type.Choice

>>> print(feature.choices)
['Not set', 'Light preset 1', 'Light preset 2', 'Light preset 3',\
 'Light preset 4', 'Light preset 5', 'Light preset 6', 'Light preset 7']
"""

from __future__ import annotations

import logging
from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from enum import Enum, auto
from functools import cached_property
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .device import Device
    from .module import Module

_LOGGER = logging.getLogger(__name__)


@dataclass
class Feature:
    """Feature defines a generic interface for device features."""

    class Type(Enum):
        """Type to help decide how to present the feature."""

        #: Sensor is an informative read-only value
        Sensor = auto()
        #: BinarySensor is a read-only boolean
        BinarySensor = auto()
        #: Switch is a boolean setting
        Switch = auto()
        #: Action triggers some action on device
        Action = auto()
        #: Number defines a numeric setting
        #: See :attr:`range_getter`, :attr:`Feature.minimum_value`,
        #: and :attr:`maximum_value`
        Number = auto()
        #: Choice defines a setting with pre-defined values
        Choice = auto()
        Unknown = -1

    # Aliases for easy access
    Sensor = Type.Sensor
    BinarySensor = Type.BinarySensor
    Switch = Type.Switch
    Action = Type.Action
    Number = Type.Number
    Choice = Type.Choice

    DEFAULT_MAX = 2**16  # Arbitrary max

    class Category(Enum):
        """Category hint to allow feature grouping."""

        #: Primary features control the device state directly.
        #: Examples include turning the device on/off, or adjusting its brightness.
        Primary = auto()
        #: Config features change device behavior without immediate state changes.
        Config = auto()
        #: Informative/sensor features deliver some potentially interesting information.
        Info = auto()
        #: Debug features deliver more verbose information then informative features.
        #: You may want to hide these per default to avoid cluttering your UI.
        Debug = auto()
        #: The default category if none is specified.
        Unset = -1

    #: Device instance required for getting and setting values
    device: Device
    #: Identifier
    id: str
    #: User-friendly short description
    name: str
    #: Type of the feature
    type: Feature.Type
    #: Callable or name of the property that allows accessing the value
    attribute_getter: str | Callable | None = None
    #: Callable coroutine or name of the method that allows changing the value
    attribute_setter: str | Callable[..., Coroutine[Any, Any, Any]] | None = None
    #: Container storing the data, this overrides 'device' for getters
    container: Device | Module | None = None
    #: Icon suggestion
    icon: str | None = None
    #: Attribute containing the name of the unit getter property.
    #: If set, this property will be used to get the *unit*.
    unit_getter: str | Callable[[], str] | None = None
    #: Category hint for downstreams
    category: Feature.Category = Category.Unset

    # Display hints offer a way suggest how the value should be shown to users
    #: Hint to help rounding the sensor values to given after-comma digits
    precision_hint: int | None = None

    #: Attribute containing the name of the range getter property.
    #: If set, this property will be used to set *minimum_value* and *maximum_value*.
    range_getter: str | Callable[[], tuple[int, int]] | None = None

    #: Attribute name of the choices getter property.
    #: If set, this property will be used to get *choices*.
    choices_getter: str | Callable[[], list[str]] | None = None

    def __post_init__(self) -> None:
        """Handle late-binding of members."""
        # Populate minimum & maximum values, if range_getter is given
        self._container = self.container if self.container is not None else self.device

        # Set the category, if unset
        if self.category is Feature.Category.Unset:
            if self.attribute_setter:
                self.category = Feature.Category.Config
            else:
                self.category = Feature.Category.Info

        if self.type in (
            Feature.Type.Sensor,
            Feature.Type.BinarySensor,
        ):
            if self.category == Feature.Category.Config:
                raise ValueError(
                    f"Invalid type for configurable feature: {self.name} ({self.id}):"
                    f" {self.type}"
                )
            elif self.attribute_setter is not None:
                raise ValueError(
                    f"Read-only feat defines attribute_setter: {self.name} ({self.id}):"
                )

    def _get_property_value(self, getter: str | Callable | None) -> Any:
        if getter is None:
            return None
        if isinstance(getter, str):
            return getattr(self._container, getter)
        if callable(getter):
            return getter()
        raise ValueError("Invalid getter: %s", getter)  # pragma: no cover

    @property
    def choices(self) -> list[str] | None:
        """List of choices."""
        return self._get_property_value(self.choices_getter)

    @property
    def unit(self) -> str | None:
        """Unit if applicable."""
        return self._get_property_value(self.unit_getter)

    @cached_property
    def range(self) -> tuple[int, int] | None:
        """Range of values if applicable."""
        return self._get_property_value(self.range_getter)

    @property
    def maximum_value(self) -> int:
        """Maximum value."""
        if range := self.range:
            return range[1]
        return self.DEFAULT_MAX

    @property
    def minimum_value(self) -> int:
        """Minimum value."""
        if range := self.range:
            return range[0]
        return 0

    @property
    def value(self) -> int | float | bool | str | Enum | None:
        """Return the current value."""
        if self.type == Feature.Type.Action:
            return "<Action>"
        if self.attribute_getter is None:
            raise ValueError("Not an action and no attribute_getter set")

        container = self.container if self.container is not None else self.device
        if callable(self.attribute_getter):
            return self.attribute_getter(container)
        return getattr(container, self.attribute_getter)

    async def set_value(self, value: int | float | bool | str | Enum | None) -> Any:
        """Set the value."""
        if self.attribute_setter is None:
            raise ValueError("Tried to set read-only feature.")
        if self.type == Feature.Type.Number:  # noqa: SIM102
            if not isinstance(value, int | float):
                raise ValueError("value must be a number")
            if value < self.minimum_value or value > self.maximum_value:
                raise ValueError(
                    f"Value {value} out of range "
                    f"[{self.minimum_value}, {self.maximum_value}]"
                )
        elif self.type == Feature.Type.Choice:  # noqa: SIM102
            if not self.choices or value not in self.choices:
                raise ValueError(
                    f"Unexpected value for {self.name}: '{value}'"
                    f" - allowed: {self.choices}"
                )

        if callable(self.attribute_setter):
            attribute_setter = self.attribute_setter
        else:
            container = self.container if self.container is not None else self.device
            attribute_setter = getattr(container, self.attribute_setter)

        if self.type == Feature.Type.Action:
            return await attribute_setter()

        return await attribute_setter(value)

    def __repr__(self) -> str:
        try:
            value = self.value
            choices = self.choices
        except Exception as ex:
            return f"Unable to read value ({self.id}): {ex}"

        if self.type == Feature.Type.Choice:
            if not isinstance(choices, list):
                _LOGGER.error(
                    "Choices are not properly defined for %s (%s). Type: <%s> Value: %s",  # noqa: E501
                    self.name,
                    self.id,
                    type(choices),
                    choices,
                )
                return f"{self.name} ({self.id}): improperly defined choice set."
            if (value not in choices) and (
                isinstance(value, Enum) and value.name not in choices
            ):
                _LOGGER.warning(
                    "Invalid value for for choice %s (%s): %s not in %s",
                    self.name,
                    self.id,
                    value,
                    choices,
                )
                return (
                    f"{self.name} ({self.id}): invalid value '{value}' not in {choices}"
                )
            value = " ".join(
                [
                    f"*{choice}*"
                    if choice == value
                    or (isinstance(value, Enum) and choice == value.name)
                    else f"{choice}"
                    for choice in choices
                ]
            )
        if self.precision_hint is not None and isinstance(value, float):
            value = round(value, self.precision_hint)

        if isinstance(value, Enum):
            value = repr(value)
        s = f"{self.name} ({self.id}): {value}"
        if (unit := self.unit) is not None:
            if isinstance(unit, Enum):
                unit = repr(unit)
            s += f" {unit}"

        if self.type == Feature.Type.Number:
            s += f" (range: {self.minimum_value}-{self.maximum_value})"

        return s
